/**************************************************************************************
    cline_traits.cxx
    code for the basic traits for the conceptual line object class

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/
#include <stdio.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_dp_graph.hxx>
#include <tml/tml_os_depends.hxx>

/* defined in node_traits.cxx where they are used by */
/*   gnode::format_line() to determine tab offsets */
extern long calc_tab_pixel_off(node*, format*, long, long);
extern long calc_tab_pixel_off_backward(node*, format*, long, long);

/**************************************************************************************
                                                                                                             
    Traits for the base class cline. This class is used for free space and blank
    space.

**************************************************************************************/

/*constructors for the base class cline*/

cline::cline(format *to_format)
{
    in_format = to_format;
    type = cl_free;
    on_phead = 0;
    next = prior = 0;
    ulx = uly = width = height = 0;
}

/*destructor for the base class cline*/
cline::~cline()
{
    in_format = NULL;
    on_phead = NULL;
    next = prior = 0;
}

/**************************************************************************************
                                                                                                             
    Traits that access conceptual line data

**************************************************************************************/

/*get the line logically below this--go to the next page if necessary*/
cline* cline::get_down()
{
    phead *next_phead;

    if (next)
        return next;

    next_phead = on_phead->next_page(1);
    return (next_phead ? next_phead->get_tline(): 0);
}
                                                                   
/*return the line logically above this--go to the prior page if necessary*/
cline* cline::get_up()
{
    phead *prior_phead;

    if (prior)
        return prior;

    prior_phead = on_phead->prior_page(1);
    return (prior_phead ? prior_phead->get_bline() : 0);
}
                                                                   


/**************************************************************************************
                                                                                                             
    Traits that do things to the conceptual line class.

**************************************************************************************/
         

/*split--splits a free conceptual line in two either horizontally or vertically.  */
/*  Argument "direction" is 0 for vertical splits, nonzero for horizontal splits. */
/*  When "from_top" is zero, creates a new cline of new_size depth (width)        */
/*    from the top (left) of "this".                                              */
/*  When "from_top" is nonzero, creates a new cline of new_size depth (width)     */
/*    from the bottom (right) of "this".                                             */
/*  Returns the space allocated.                                                  */
cline* cline::split(short new_size, short from_top , short direction )
{       
    register cline* new_cline;

    /*can't split a line which isn't free space*/
    if (type != cl_free)
        return 0;
                                
    /*check to make sure there is enough space to do the split*/
    if (direction) {
        if (width < new_size)
            return 0;
        }
    else if (height < new_size)
        return 0;

    new_cline = new cline;
    if (!new_cline) 
        return new_cline;

    *new_cline = *this;

    switch (from_top){
      case 0:
        if (prior)
            prior->next = new_cline;
        new_cline->next = this;
        new_cline->prior = prior;
        prior = new_cline;
        if (on_phead->top_line == this)
            on_phead->top_line = new_cline;
        if (on_phead->top_free == this)
            on_phead->top_free = new_cline;
        switch (direction) {
          case 0:
            uly += new_size;
            height -= new_size;
            new_cline->height = new_size;
            break;

          default:
            ulx += new_size;
            width -= new_size;
            new_cline->width = new_size;
            break;
          }
        break;

      default:
        if (next)
            next->prior = new_cline;
        new_cline->prior = this;
        new_cline->next = next;
        next = new_cline;
        if (on_phead->bottom_line == this)
            on_phead->bottom_line = new_cline;
        if (on_phead->bottom_free == this)
            on_phead->bottom_free = new_cline;
        switch (direction) {
          case 0:
            height -= new_size;
            new_cline->uly += height;
            new_cline->height = new_size;
            break;

          default:
            width -= new_size;
            new_cline->ulx += width;
            new_cline->width = new_size;
            break;
          }
        break;
      }

    return new_cline;
}

/*resizes the height of a conceptual line and all lines below the adjusted line--NOTE--this is a rather simple minded
    implementation--it will have to be made smarter if we do multiple columns or wrapping around
    graphic objects*/
long cline::resize(short hi_delta)
{
    register cline *next_line,
                   *hold_line;
    
    /*if there isn't enough room on the conceptual page, return 0*/
    if ((uly + height + hi_delta) > (on_phead->get_hi())) return 0;

    /*update this conceptual line, and any lines on the same y_axis*/
    height += hi_delta;
    for (next_line = prior; next_line && (next_line->uly == uly); next_line = next_line->prior)
        next_line->height += hi_delta;
    for (next_line = next; next_line && (next_line->uly == uly); next_line = next_line->next)
        next_line->height += hi_delta;

    /*update all lines below the one resized*/
    for (next_line = next; next_line; next_line = next_line->next) {

        if (next_line->type != cl_free)     //space is used--just adjust the coordinates
            next_line->set_ul_coord(next_line->ulx, next_line->uly + hi_delta);
        else {         //space is free--see if we can eat the change in this space
            if (next_line->height > hi_delta) {
                next_line->uly -= hi_delta;
                next_line->height -= hi_delta;
                return 1;
                }
            else
                next_line->set_ul_coord(next_line->ulx, next_line->uly + hi_delta);

            if (next_line->uly > on_phead->get_hi()){  //we're off the page
                /*first, find the previous free space, if there is one*/
                for (hold_line = next_line->prior; hold_line && hold_line->type != cl_free; hold_line = hold_line->prior)
                    ;
                on_phead->bottom_free = hold_line;
                if (!hold_line) on_phead->top_free = hold_line;

                /*now, find the last line with something in it and mark the page head with the new end_pos*/
                for (hold_line = next_line->prior; hold_line && (0 == hold_line->get_start().in_node); hold_line = hold_line->prior)
                    ;
                if (hold_line) {
                    on_phead->set_end_pos(hold_line->get_end());
                    if (hold_line->get_phead()->get_next()) {
                        hold_line->get_phead()->get_next()->set_start_pos(hold_line->get_end().get_next());
                        hold_line->get_phead()->get_next()->mark_reformat();
                        }
                    hold_line->get_phead()->free_clines(hold_line);
                    }
                /*else nothing formattable from the model was on the page so leave end_pos of phead as it was*/
                /* this should cover the case of a phead containing only clines of types cl_blank, cl_gutter, cl_other */
                }
            }
        }
    return 1;
}

/*attach a line before (or after) this line--needs to update the coordinates of the
    lines below*/
void cline::attach(cline *target_line, long before)
{
    on_phead = target_line->on_phead;
    if (before){
        set_prior(target_line->prior);
        if (on_phead->get_start() >= get_start())
            on_phead->set_start_pos(get_start());
        if (prior)
            prior->set_next(this);
        else {
            on_phead->set_top_line(this);
            if (type == cl_free)
                on_phead->set_top_free(this);
            else {
                on_phead->set_top_used(this);
                }
            }
        target_line->set_prior(this);
        set_next(target_line);
        set_ul_coord(target_line->ulx, target_line->uly);
        on_phead->adjust_xy(target_line, 0, height);
        }
    else {
        set_next(target_line->next);
        if (on_phead->get_end() <= get_end()){
            on_phead->set_end_pos(get_end());
            if (on_phead->get_next()){
                on_phead->get_next()->set_start_pos(get_end().get_next());
                on_phead->get_next()->mark_reformat();
                }
            }
        if (next)
            next->set_prior(this);
        else {
            on_phead->set_bottom_line(this);
            if (type == cl_free){
                on_phead->set_bottom_free(this);
                }
            else {
                on_phead->set_bottom_used(this);
                }
            }
        target_line->set_next(this);
        set_prior(target_line);
        set_ul_coord(target_line->get_ulx(), target_line->get_uly() +target_line->get_hi());
        on_phead->adjust_xy(next, 0, height);
        }
}

/*detach a line from a conceptual page--needs to update the coordinates of the lines
    below*/
void cline::detach()
{
    if (!on_phead) return;

    if (prior)
        prior->set_next(next);
    else {
        on_phead->set_top_line(next);
        if (next->type == cl_free)
            on_phead->set_top_free(next);
        else
            on_phead->set_top_used(next);
        }

    if (next) {
        next->set_prior(prior);
        on_phead->adjust_xy(next, 0, 0 - height);
        release();
        }
    else {
        if (prior->type == cl_free) {
            prior->resize(height);
            on_phead->set_bottom_free(prior);
            on_phead->set_bottom_line(prior);
            release();
            }
        else {
            set_type(cl_free);
            prior->set_next(this);
            }
        }
}



/**************************************************************************************
                                                                                                             
    Implementations of the virtual functions for base class cline

**************************************************************************************/
                                                                                       
/*get the next line--go to the next page if necessary*/
cline* cline::get_next()
{
    phead *next_phead;
    if (next)
        return next;

    next_phead = on_phead->next_page(1);
    if (!next_phead)
        return 0;
    return next_phead->get_tline();
}
                                                                   
/*return the prior conceptual line*/
cline* cline::get_prior()
{
    phead     *prior_phead;
    cline     *return_line;
    model_pos  find_pos;

    if (prior)
        return prior;

    find_pos = on_phead->get_start();
    prior_phead = on_phead->prior_page(1);
    if (!prior_phead)
        return 0;
    if (prior_phead->get_end() > find_pos){
        find_pos = find_pos.get_prior();
        return_line = prior_phead->mpos_2_hcline(find_pos);
        if (return_line)
            return return_line;
        }
    return prior_phead->get_bline();
}

/*get the model position which starts the line--always returns 0,-1 for clines*/
model_pos cline::get_start()
{
    model_pos return_pos;

    return_pos.in_node = 0;
    return_pos.char_index = -1;
    return return_pos;
}

/*the the model position which ends the line--always returns 0, -1 for clines*/
model_pos cline::get_end()
{
    model_pos return_pos;

    return_pos.in_node = 0;
    return_pos.char_index = -1;
    return return_pos;
}

/*set the x_offset to the beginning of the drawable--a noop in the base case*/
void cline::set_start_x_off(long new_x_off)
{
    return;
}

/*get the x_offset to the beginning of the drawable--a noop in the base case*/
long cline::get_start_x_off()
{
    return 0;
}

/*does the line contain the specified model position? Clines don't contain data*/
long cline::contains(model_pos target_pos)
{
    return 0;
}

/*replace the indicated line with this. Gets rid of the line pointed to*/
void cline::replace(cline *replace_line)
{
    register cline *hold_cline;

    set_ul_coord(replace_line->ulx, replace_line->uly);
    next = replace_line->next;
    if (next)
        next->prior = this;
    prior = replace_line->prior;
    if (prior)
        prior->next = this;
    replace_line->prior = replace_line->next = 0;
    on_phead = replace_line->get_phead();
    if (replace_line == on_phead->top_line)
        on_phead->top_line = this;
    if (replace_line == on_phead->bottom_line)
        on_phead->bottom_line = this;
    if (replace_line == on_phead->top_used)
        on_phead->top_used = this;
    if (replace_line == on_phead->bottom_used)
        on_phead->bottom_used = this;
    if (replace_line == on_phead->top_free){
        if (type == cl_free)
            on_phead->top_free = this;
        else {
            for (hold_cline = inq_next(); hold_cline && (hold_cline->type != cl_free); hold_cline = hold_cline->inq_next())
                ;
            on_phead->top_free = hold_cline;
            }
        }
    if (replace_line == on_phead->bottom_free){
        if (type == cl_free)
            on_phead->bottom_free = this;
        else {
            for (hold_cline = inq_prior(); hold_cline && (hold_cline->type != cl_free); hold_cline = hold_cline->inq_prior())
                ;
            on_phead->bottom_free = hold_cline;
            }
        }
    delete replace_line;
}

/*allocate space from this line and fill it with the line supplied.               */
/*  Return 0 if this is not free space or this is smaller than line supplied.     */
/*  Otherwise:                                                                    */
/*  - if at_bottom = 0, allocate space from the top (left) of this line.          */
/*  - if at_bottom is nonzero, put fill_with at the bottom (right) of the line.   */
/*  - if space remains in this line after the allocation, return it,              */
/*      otherwise return closest free line to this in the indicated direction.    */
cline* cline::alloc_space(cline *fill_with, long at_bottom )
{
    register cline *return_line,
                   *split_line,
                   *new_line;
             tline *hold_tline;
             long   last_alloced;

    if (type != cl_free)
        return 0;

    if (fill_with->width > width)
        return 0;

    if (fill_with->height > height)
        return 0;
                                                                                
    if (cl_text == fill_with->get_type()) {
        for (hold_tline = (tline*) fill_with; hold_tline; hold_tline = (tline*) hold_tline->get_right())
            hold_tline->set_phead(on_phead);
        }
    else
        fill_with->on_phead = on_phead;
    if (fill_with->width == width) {         //no splitting necessary!!
        if (fill_with->height == height) {   //just replace
            if (at_bottom)
                last_alloced = (this == on_phead->top_free);
            else
                last_alloced = (this == on_phead->bottom_free);
            fill_with->cline::replace(this);
            if (at_bottom) {
                for (return_line = fill_with->prior;
                  return_line && (return_line->type != cl_free);
                  return_line = return_line->prior)
                   ;
                }
            else {
                for (return_line = fill_with->next;
                  return_line && (return_line->type != cl_free);
                  return_line = return_line->next)
                   ;
                }
            if (at_bottom)
                return (last_alloced) ? fill_with->inq_next()  : return_line;
            else
                return (last_alloced) ? fill_with->inq_prior() : return_line;
            }
        if (at_bottom){
            height -= fill_with->height;
            fill_with->set_ul_coord(ulx, uly + height);
            fill_with->next = next;
            fill_with->prior = this;
            if (next)
                next->prior = fill_with;
            else if (on_phead) 
                on_phead->bottom_line = fill_with;
            next = fill_with;
            return this;
            }
        else {
            height -= fill_with->height;
            fill_with->set_ul_coord(ulx, uly);
            uly += fill_with->height;
            fill_with->prior = prior;
            fill_with->next = this;
            if (prior)
                prior->next = fill_with;
            else if (on_phead) 
                on_phead->top_line = fill_with;
            prior = fill_with;
            return this;
            }
        }
    
    /* split vertically */
    new_line = split((short) fill_with->height, (short) at_bottom);

    /* split horizontally */
    split_line = new_line->split(fill_with->width, (short) 0, (short) 1);

    fill_with->replace(split_line);

    return this;
}

/*return the xy_coord of the indicated model position. Note that in the base
    class of cline no model position can be contained, so the invalid xy_coord
    (-1,-1) is always returned*/
xy_coord cline::mpos_2_xy(model_pos find_pos)
{
    xy_coord return_xy;
    
    return_xy.x_coord = -1;
    return_xy.y_coord = -1;
    return return_xy;
}

/*return the model_pos associated with an xy_coord. Note that since no
    model position can be contained in a base cline, the model pos returned
    is always the invalid position in_node = 0, char_index = -1*/
model_pos cline::xy_2_mpos(xy_coord where)
{
    model_pos return_pos;

    return_pos.in_node = 0;
    return_pos.char_index = -1;
    return return_pos;
}

/*draw the area for the cline, and return a view_pos if the area is not free.
    if the area is free, return 0*/
view_pos* cline::draw_line(view_port *in_viewport, screen_xy where, long start_index)
{
    register view_pos *return_pos;

    if ((type == cl_free) || (type == cl_blank)) return 0;

    return_pos = in_viewport->get_pos();
    if (!return_pos) return 0;

    return_pos->set_x(where.at_x);
    return_pos->set_y(where.at_y);
    return_pos->set_line(this);
    
    return return_pos;
}

/*reformat the line--should always be done by one of the subclass functions*/
cline* cline::reformat(model_pos new_start, long& next_ok, long force_start )
{
             model_pos  next_start;
    register cline     *new_line;
    
    next_ok = 0;
    new_line = new_start.in_node->format_line(in_format, new_start.char_index, on_phead->get_hi() - uly,
                on_phead->get_wide() - ulx);
    if (new_line)
        if (!next || (next->get_type() == cl_free) || (next->get_start() == new_line->get_end().get_next()))
            next_ok = 1;
    return new_line;
}


/**************************************************************************************
                                                                                                             
    Implementations of the derived class tline from base class cline

**************************************************************************************/
                                                                                       
/*constructor for the class tline*/
tline::tline(format *to_format ) : (to_format)
{
    right = left = 0;
    use_font = 0;
    start_pos.in_node = 0;
    start_pos.char_index = 0;
    count = 0;
    baseline_y = text_width = 0;
}
                                                        
/*destructor for the class tline*/
tline::~tline()
{
    right = left = 0;
    use_font = 0;
    start_pos.in_node = 0;
}

/*adjust the baseline of this tline, and all of the baselines for the tlines to the right*/
long tline::adjust_baseline(long new_baseline_y)
{
    register tline *do_tline;

    for (do_tline = this; do_tline; do_tline = do_tline->right)
        baseline_y = (short) new_baseline_y;

    return new_baseline_y;
}
                         
/*return the width of this text line and all of it's right siblings*/
long tline::get_full_text_width()
{
    register long   return_width;
    register tline *this_line;

    return_width = 0;
    for (this_line = this; this_line; this_line = this_line->right)
        return_width += this_line->text_width + this_line->start_x_offset;

    return return_width;
}


void tline::set_right(tline *new_right)
{
    if (right){
        right->left = new_right;
        new_right->right = right;
        }

    new_right ->left = this;
    right = new_right;
}

void tline::set_left(tline *new_left)
{
    if (left){
        left->right = new_left;
        new_left->left = left;
        }

    new_left->right = this;
    left = new_left;
}
            
long tline::trunc_line (long max_x)
{
   register xy_coord  last_xy;
            model_pos last_pos = start_pos;
                         
   last_pos.char_index += count - 1;

   last_xy = mpos_2_xy (last_pos);

   if (last_xy.x_coord > max_x)
   {
      last_xy.x_coord = max_x;

      last_pos = xy_2_mpos (last_xy);

      return (last_pos.in_node == start_pos.in_node)
             ? last_pos.char_index - start_pos.char_index
             : (last_pos < start_pos) ? 0 : count;
   }
   else
      return count;
}

/**************************************************************************************
                                                                                                             
    Implementations of the derived class tline virtual functions

**************************************************************************************/
/*get the next line--if there is a right pointer, return it; else go back to the
    first tline in the left-to-right list and return its down pointer*/
cline* tline::get_next()
{
    register tline* hold_line;

    if (right) return right;

    for (hold_line = this; hold_line->left; hold_line = hold_line->left)
        ;
    return hold_line->cline::get_next();
}

/*set the next pointer. If there is enough space on the page (in width) to
    contain the line, attach it to the right; else attach it down using the
    parent class function*/
void tline::set_next(cline *next_line)
{
    register tline *hold_line;

    for (hold_line = this; hold_line; hold_line = hold_line->right)
        hold_line->cline::set_next(next_line);
}

/*set the prior pointer. If there is enough space on the page (in width) to
    contain the line, attach it to the right; else attach it up using the
    parent class function*/
void tline::set_prior(cline *prior_line)
{
    register tline *hold_line;

    for (hold_line = this; hold_line; hold_line = hold_line->right)
        hold_line->cline::set_prior(prior_line);
}

/*get the model position of the last character on the line*/
model_pos tline::get_end()
{
    register tline     *next_line;
             model_pos  return_pos;
             char_set  *in_char_set;

    for (next_line = this; next_line->right; next_line = next_line->right)
        ;
    return_pos = next_line->start_pos;
    in_char_set = (char_set*) return_pos.in_node->get_model()->inq_state(return_pos.in_node, tml_char_set_prop);
    return_pos.char_index += (next_line->count - in_char_set->get_char_size());
    return return_pos;
}

/*set the ul coordinates of the line, and all of its right siblings*/
void tline::set_ul_coord(short upper_x, short upper_y)
{
    cline::set_ul_coord(upper_x, upper_y);
    if (right)
        right->set_ul_coord((short) (upper_x + text_width + start_x_offset), (short) upper_y);
}

/*adjust the ulx and uly coordinates by the indicated amount for all the lines on this y*/
void tline::adjust_xy(long deltax, long deltay)
{
    register tline *adjust_line;

    for (adjust_line = this; adjust_line->left; adjust_line = adjust_line->left)
        ;
    /* tline::set_ul_coord() calls itself recursively on right siblings */
    adjust_line->set_ul_coord((short) (adjust_line->get_ulx() + deltax), (short) (adjust_line->get_uly() + deltay));
}

       
/*does the tline or any of it's siblings contain the indicated model position?*/
long tline::contains(model_pos target_pos)
{
    register tline *on_tline;

    for (on_tline = this; on_tline; on_tline = on_tline->right){
        if ((on_tline->start_pos.in_node == target_pos.in_node)
            &&(on_tline->start_pos.char_index <= target_pos.char_index)
            && ((on_tline->start_pos.char_index + count) > target_pos.char_index))
            return 1;
        }
    return 0;
}

/*replace a line with this tline--take the old line out of the list of lines,
    put this line in its place, and then adjust the next and prior pointers
    of all of the siblings*/
void tline::replace(cline *replace_line)
{
    register tline *first_line,
                   *hold_line,
                   *next_line;
    register cline *new_replace;

    for (first_line = this; first_line->left; first_line = first_line->left)
        ;
    if (replace_line->get_type() == cl_text){
        next_line = (tline *)replace_line;
        first_line->left = next_line->left;
        for (hold_line = first_line; hold_line; hold_line = hold_line->right)
            hold_line->set_phead(next_line->get_phead());
        if (first_line->left){
            first_line->set_ul_coord(first_line->left->get_lrx(), first_line->left->get_uly());
            first_line->left->right = first_line;
            }
        else {
            first_line->set_ul_coord(replace_line->get_ulx(), replace_line->get_uly());
            }
        first_line->set_prior(next_line->get_up());
        first_line->set_next(next_line->get_down());
        if (first_line->get_up())
            first_line->get_up()->set_next(first_line);
        else {
            first_line->get_phead()->set_top_used(first_line);
            if (first_line->get_phead()->get_tline() == next_line)
                first_line->get_phead()->set_top_line(first_line);
            }
        if (first_line->get_down())
            first_line->get_down()->set_prior(first_line);
        else {
            first_line->get_phead()->set_bottom_used(first_line);
            if (first_line->get_phead()->get_bline() == next_line)
                first_line->get_phead()->set_bottom_used(first_line);
            }
        for (hold_line = next_line; hold_line; hold_line = hold_line->right){
            hold_line->set_next(0);
            hold_line->set_prior(0);
            }
        replace_line->release();
        }
    else{
        if (replace_line->get_type() == cl_free){
            new_replace = replace_line->alloc_space(first_line);
            if (!new_replace)
                first_line->cline::replace(replace_line);
            }
        else
            first_line->cline::replace(replace_line);
        }

}
                         
void tline::release()
{
    register tline *hold_line,
                   *next_line;

    if (left) {
        left->right = NULL;
        for (hold_line = this; hold_line; hold_line = next_line) {
            next_line = hold_line->right;
            delete hold_line;
            }
        }
    else if (right) {
        /* head tline released by base class call */
        right->release();
        right = NULL;
        }

    cline::release();
}

/*return the conceptual page coordinates of the indicated character. If the character is not
    on this group of conceptual lines, return -1,-1*/
xy_coord tline::mpos_2_xy(model_pos find_pos)
{
             xy_coord  return_xy;
    register char     *line_st;
    register long      i,
                       char_size,
                       end_index,
                       tab_pixel_off,
                       tab_cnt;
             long     *margin_array;
             long     *format_value;
    register char_set *in_char_set = NULL;
             long      char_val, space_remaining;
             lrud_dir  text_dir;

    if (!find_pos.in_node || !start_pos.in_node){
        return_xy.x_coord = -1;
        return_xy.y_coord = -1;
        return return_xy;
        }
    if (find_pos < start_pos) {
        return_xy.x_coord = -1;
        return_xy.y_coord = -1;
        return return_xy;
        }
        
    if ((find_pos.in_node != start_pos.in_node) || (find_pos.char_index >= start_pos.char_index + count))
        if (right)
            return right->mpos_2_xy(find_pos);
        else {
            return_xy.x_coord = -1;
            return_xy.y_coord = -1;
            return return_xy;
            }
                              
    if (start_pos.in_node->get_model())
        in_char_set = (char_set *)start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_char_set_prop);
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    return_xy.y_coord = get_uly() + baseline_y;
    return_xy.x_coord = get_ulx();
    space_remaining = get_text_width() + get_start_x_off();
    line_st = ((char*) start_pos.in_node->get_contents()) + start_pos.char_index;
    end_index = start_pos.char_index + count;
    if (::right == (text_dir = in_char_set->get_dir_to_char((unsigned char*)line_st))) {
        if (!left && start_pos.in_node->get_model()){
            margin_array = (long *)(start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_margin_prop));
            if (margin_array) {
                return_xy.x_coord += margin_array[left_marg];
                space_remaining -= margin_array[left_marg];
                }
            }
        for (tab_cnt = 0, i = start_pos.char_index; 
          (i < end_index) && (i < find_pos.char_index) && in_char_set->is_htab((unsigned char*)line_st);
          line_st += char_size, tab_cnt++, i += char_size)
            ;
        if (tab_cnt) {
            if (-1 != (tab_pixel_off = calc_tab_pixel_off(start_pos.in_node, get_format(), tab_cnt, return_xy.x_coord)))
                return_xy.x_coord += tab_pixel_off;
            else
                return_xy.x_coord += space_remaining;
            }
        }
    else if (::left == text_dir) {
        return_xy.x_coord += (get_start_x_off() + get_text_width());
        if (!right && start_pos.in_node->get_model()){
            margin_array = (long *)(start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_margin_prop));
            if (margin_array) {
                return_xy.x_coord -= margin_array[right_marg];
                space_remaining -= margin_array[right_marg];
                }
            }
        for (tab_cnt = 0, i = start_pos.char_index;
          (i < end_index) && (i < find_pos.char_index) && in_char_set->is_htab((unsigned char*) line_st);
          line_st += char_size, tab_cnt++, i += char_size)
            ;
        if (tab_cnt) {
            if (-1 != (tab_pixel_off = calc_tab_pixel_off_backward(start_pos.in_node, get_format(), tab_cnt, return_xy.x_coord)))
                return_xy.x_coord -= tab_pixel_off;
            else
                return_xy.x_coord -= space_remaining;
            }
        }

    if ((0 == tab_cnt)
      && (format_value = (long*) get_format()->inq_state(start_pos.in_node, tml_wrap_prop))
      && (word_wrap == *format_value)
      && (((::right == text_dir) && (0 == get_ulx()))
         || 
         ((::left == text_dir) && ((get_ulx() + get_wide()) == get_phead()->get_wide()))
         || 
         (((::down == text_dir) /*|| (::stack_down == text_dir)*/) && (0 == get_uly()))
         ||
         (((::up == text_dir)   /*|| (::stack_up == text_dir)*/) && ((get_uly() + get_hi()) == get_phead()->get_hi())))) {
        for (i = start_pos.char_index; 
          i < end_index; 
          i += char_size, line_st += char_size) {
            if ( ! in_char_set->cs_isspace((unsigned char*) line_st))
                break;
            }
        }

    for (; 
      (i < end_index) && (i < find_pos.char_index); 
      i += char_size, line_st += char_size) {
        char_val = in_char_set->as_ulong((unsigned char*) line_st);
        if (::right == text_dir)
            return_xy.x_coord += use_font->get_width(char_val) + use_font->get_icspace();
        else if (::left == text_dir)
            return_xy.x_coord -= use_font->get_width(char_val) + use_font->get_icspace();
        }
    return return_xy;
}

/*return the model position of the character at the conceptual page coordinate specified.
    If the y coordinate specified is outside the range of the line, return 0,-1; if the x_coordinate
    is outside the range specified, either check the right tline or return the end_pos*/
model_pos tline::xy_2_mpos(xy_coord where)
{
    register char      *line_st;
    register long       i,
                        icspace,
                        char_size,
                        last_index,
                        tab_pixel_off,
                        tab_cnt;
    register char_set  *in_char_set = NULL;
             model_pos  return_pos;
             long      *margin_array;
             long      *format_value;
             xy_coord   current_xy;
             long       char_val, space_remaining;
             lrud_dir   text_dir;

    if ((where.y_coord < get_uly()) || (where.y_coord > get_lry())) {
        return_pos.in_node = 0;
        return_pos.char_index = -1;
        return return_pos;
        }

    if (where.x_coord < get_ulx()) {
        if (left)
            return left->xy_2_mpos(where);
        else
            return start_pos;
        }

    if (where.x_coord >= (get_ulx() + start_x_offset + text_width)) {
        if (right)
            return right->xy_2_mpos(where);
        else {
            return_pos.in_node = start_pos.in_node;
            return_pos.char_index = start_pos.char_index + count - 1;
            return return_pos;
            }
        }

    if (start_pos.in_node->get_model())
        in_char_set = (char_set *)start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_char_set_prop);
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    current_xy.x_coord = get_ulx();
    space_remaining = get_text_width() + get_start_x_off();
    line_st = ((char*) start_pos.in_node->get_contents()) + start_pos.char_index;
    last_index = start_pos.char_index + count - char_size;

    if (::right == (text_dir = in_char_set->get_dir_to_char((unsigned char*)line_st))) {
        if (!left && start_pos.in_node->get_model()) {
            margin_array = (long *)(start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_margin_prop));
            if (margin_array) {
                current_xy.x_coord += margin_array[left_marg];
                space_remaining -= margin_array[left_marg];
                }
            }
        for (tab_cnt = 0, i = start_pos.char_index;
          (i < last_index) && in_char_set->is_htab((unsigned char*) line_st);
          line_st += char_size, tab_cnt++, i += char_size)
            ;
        if (tab_cnt) {
            if (-1 != (tab_pixel_off = calc_tab_pixel_off(start_pos.in_node, get_format(), tab_cnt, current_xy.x_coord)))
                current_xy.x_coord += tab_pixel_off;
            else
                current_xy.x_coord += space_remaining;
            }
        }
    else if (::left == text_dir) {
        current_xy.x_coord += get_text_width() + get_start_x_off();
        if (!right && start_pos.in_node->get_model()) {
            margin_array = (long *)(start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_margin_prop));
            if (margin_array) {
                current_xy.x_coord -= margin_array[right_marg];
                space_remaining -= margin_array[right_marg];
                }
            }
        for (tab_cnt = 0, i = start_pos.char_index; 
          (i < last_index) && in_char_set->is_htab((unsigned char*) line_st);
          line_st += char_size, tab_cnt++, i += char_size)
            ;
        if (tab_cnt) {
            if (-1 != (tab_pixel_off = calc_tab_pixel_off_backward(start_pos.in_node, get_format(), tab_cnt, current_xy.x_coord)))
                current_xy.x_coord -= tab_pixel_off;
            else
                current_xy.x_coord -= space_remaining;
            }
        }

    if ((0 == tab_cnt)
      && (0 == get_ulx())
      && (format_value = (long*) get_format()->inq_state(start_pos.in_node, tml_wrap_prop))
      && (word_wrap == *format_value)
      && (((::right == text_dir) && (0 == get_ulx()))
         || 
         ((::left == text_dir) && ((get_ulx() + get_wide()) == get_phead()->get_wide()))
         || 
         (((::down == text_dir) /*|| (::stack_down == text_dir)*/) && (0 == get_uly()))
         ||
         (((::up == text_dir)   /*|| (::stack_up == text_dir)*/) && ((get_uly() + get_hi()) == get_phead()->get_hi())))) {
        for (i = start_pos.char_index; 
          i < last_index; 
          i += char_size, line_st += char_size) {
            if ( ! in_char_set->cs_isspace((unsigned char*) line_st))
                break;
            }
        }

    where.x_coord -= current_xy.x_coord;
    icspace = use_font->get_icspace();
    for (;
      char_val = in_char_set->as_ulong((unsigned char*) line_st),
      (i < last_index) && (0 <= (where.x_coord -= use_font->get_width(char_val)));
      i += char_size, line_st += char_size)
        where.x_coord -= icspace;
        ;
                                                                               
    return_pos.in_node = start_pos.in_node;
    return_pos.char_index = i;
    return return_pos;
}
                                 
/* reformat the line */
/* called indirectly through virtual function mechanism */
/*   on head clines of type cl_text */
/*   by view_port::redraw(model_pos, model_pos, view_pos) */
cline* tline::reformat(model_pos new_start, long& next_ok, long force_start )
{
    register long       open_width,
                        open_height,
                        old_max_width,
                        index_inc,
                        depth;
             long      *margin_array,
                        right_margin,
                        left_margin,
                        index_diff,
                        overflow_y;
             model_pos  old_next,
                        old_start,
                        old_end,
                        new_end,
                        new_next,
                        new_page_start,
                        next_page_start,
                        hold_start,
                        hold_end,
                        prev_end;
    register cline     *new_cline,
                       *return_line,
                       *adjust_start,
                       *hold_cline,
                       *prev_cline,
                       *next_cline,
                       *overflow,
                       *new_bottom_line; 
             cline     *last_line;
    register tline     *start_line,
                       *hold_line,
                       *new_tline;
             phead     *old_page,
                       *prev_page,
                       *next_page,
                       *page_after_overflow;
             node      *hold_node,
                       *old_node;
             char_set  *in_char_set;
             lrud_dir   text_dir;

    /* remember maximum line width for page (old_max_width) */
    /* zero value signals this line not longest on page */
    for(hold_line = this; hold_line->left; hold_line = hold_line->left)
        ;
    if ((hold_line->get_contents_width() + hold_line->get_start_x_off())
      == get_phead()->get_max_line_width())
        old_max_width = get_phead()->get_max_line_width();
    else
        old_max_width = 0;
                           
    /* remember width and maximum possible height of line */
    open_width = hold_line->get_wide();
    open_height = get_phead()->get_hi();

    /* correct and remember interesting model positions */
    /* correction required where model position refers to text */
    /*   in rightmost of three nodes created from a call to */
    /*   gnode::insert() */
    /* correction is unneeded and has no effect when no split has occurred */
    /*   but test to determine need for correction performs more work */
    /*   than correction itself so test not made */

    /* remember model position (old_start) for beginning of line to be reformatted (this) */
    old_start = start_pos;

    /* remember last model position that can be affected by this call */
    /* do this before normalizing start_pos by assigning normalized old_start to it */
    /*   because get_end() uses start_pos and character count to get end pos */
    old_end = get_end();

    /* correct old_start */
    old_start.normalize();

    /* correct start_pos member of "this" */
    /*   thereby preventing failure of gnode::format_line() */
    start_pos = old_start;

    /* remember model position (old_next) for beginning of next nonfree cline (new_cline) */
    /* do not normalize old_next as it is used to detect unnormalized model positions */
    /*   that must be corrected in tlines created from material that has since become */
    /*   the right sibling node of a node split */
    for (new_cline = inq_next();
      new_cline && (new_cline->get_type() == cl_free);
      new_cline = new_cline->inq_next())
        ;
    if (new_cline)
        old_next = new_cline->get_start();
    else if (get_phead()->get_next())
        old_next = get_phead()->get_next()->get_start();
    else {
        old_next.in_node = 0;
        old_next.char_index = -1;
        }

    /* if start position of line is to change */
    /*   update start_pos member of "this" and start member of associated phead */
    if (force_start
      /* new start forced when incremental view_port::redraw() notices */
      /*   a request to start reformatting at a position less than or */
      /*   equal to the last valid model position in the line that is to be */
      /*   reformatted */
      || (new_start < start_pos)
      /* caller specified start position for reformatting less than current line start */
      || (start_pos.in_node->get_max_index() < start_pos.char_index)) {
      /* node in which this line starts has been split */
        start_pos = new_start;
        if (get_phead()->get_start() > new_start)
            get_phead()->set_start_pos(new_start);
        }

    /* get right and left margin offsets */
    margin_array = (long*) get_format()->inq_state(start_pos.in_node, tml_margin_prop);
    if (margin_array){
        right_margin = margin_array[right_marg];
        left_margin = margin_array[left_marg];
        }
    else
        right_margin = left_margin = 0;

    /* correct width available for formatting */
    open_width -= (right_margin + left_margin);

    /* reformat this line */
    in_char_set = (char_set*) start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_char_set_prop);
    if (::right == (text_dir = in_char_set->get_dir_to_char(((unsigned char*) start_pos.in_node->get_contents()) + start_pos.char_index)))
        new_cline = start_pos.in_node->format_line(get_format(),
                                                   start_pos.char_index,
                                                   open_height,
                                                   open_width);
    else if (::left == text_dir)
        new_cline = start_pos.in_node->tamrof_line(get_format(),
                                                   start_pos.char_index,
                                                   open_height,
                                                   open_width);

    /* could not reformat line */
    /* return NULL and indicate next line needs reformatting */
    if ( ! new_cline) {
        next_ok = 0;
        return NULL;
        }

    /* correct width members of reformatted line */
    new_cline->set_start_x_off(new_cline->get_start_x_off() + left_margin);
    new_cline->set_wide((short) (new_cline->get_wide() + (left_margin + right_margin)));

    /* correct maximum line width for page */
    open_width = new_cline->get_contents_width() + left_margin + right_margin;
    if (open_width > get_phead()->get_max_line_width())
        get_phead()->set_max_line_width(open_width);
    else if (old_max_width && (old_max_width != open_width))
        get_phead()->calc_max_line_width();
                                                                           
    /* reformatted line not a tline */
    /* insert it on conceptual page before "this" */
    /* return it and indicate next line (== "this") needs no reformatting */
    if (new_cline->get_type() != cl_text){
        next_ok = 1;
        new_cline->attach(this, 1);
        return new_cline;
        }

    /* remember end of newly formatted line */
    /*   and beginning of next line to follow it */
    new_tline = (tline*) new_cline;
    new_end = new_cline->get_end();
    new_next = new_end.get_next();

    /* next character after new line is first character of old line */
    /* insert new line on conceptual page before "this" */
    if (new_next == old_start) {
        /* correct start_pos of "this" to beginning of old next line */
        /* indicate via old_next that "this" follows new line */
        /* show that "this" will begin with what used to be */
        /*   the position following it */
        start_pos = old_next = old_start;
        /* indicate first line whose start position may need updating (== this) */
        adjust_start = this;
        /* indicate next line (== this) needs no reformatting */
        next_ok = 1;
        /* sets upper left and upper right of right siblings, too */
        new_cline->set_ul_coord(get_ulx(), get_uly());
        /* insert new line before "this" */
        new_cline->attach(this, 1);
        /* cline::attach() calls phead::adjust_xy() which marks page as needing reformatting */
        /*   in case of overflow */
        /* since this function is called only in the context of incremental reformatting, */
        /*   and since overflow for this page is handled below, unmark the page */
        get_phead()->format_ok = 1;
        /* set page head members of right siblings */
        for (hold_line = new_tline; hold_line; hold_line = hold_line->right)
            hold_line->set_phead(get_phead());
        /* increment line count of page head */
        get_phead()->adjust_line_cnt(1);
        /* check for insertion above first line */
        /*   update phead appropriately */
        /* no need to check that these are not reset in adjustment loop at bottom */
        /*   of this function: since page start is set from line start and */
        /*   line start has been normalized by gnode::format_line, they will not be reset */
        for (hold_cline = this; hold_cline->inq_prior(); hold_cline = hold_cline->inq_prior())
            ;
        if (this == hold_cline) {
            get_phead()->set_top_line(new_tline);
            get_phead()->set_start_pos(new_tline->get_start());
            }
        /* prepare to return reformatted line */
        return_line = new_cline;
        }
    /* next character after new line is not first character of old line */
    /* reuse "this" to contain reformatted material */
    else {
        /* indicate first line whose start position may need updating */
        if (NULL == (adjust_start = inq_next())) {
            /* "this" is last line on page */
            for (prev_end = new_end, prev_page = get_phead(), next_page = prev_page->get_next();
              next_page;
              prev_end = next_page->get_end(), prev_page = next_page, next_page = prev_page->get_next()) {
                /* set start_pos of next page correctly in case call to phead::get_tused() */
                /*   reformats page (will happen if page format is not OK and page is not visible) */
                /* if page is reformatted, adjust_start will be moved to beginning of next page */
                /*   by next iteration of loop, or will be set to NULL if there is no next page */
                /* since adjust_start is moved beyond page, no need to check that start_pos of page */
                /*   is not reset in adjustment loop at bottom of this function */
                next_page->set_start_pos(prev_end.get_next());
                /* NOTE: get_tused() may cause next page to be reformatted */
                for (hold_cline = prev_page->get_next()->get_tused();
                  hold_cline->inq_next() && (cl_text != hold_cline->get_type());
                  hold_cline = hold_cline->inq_next())
                    ;
                adjust_start = hold_cline;
                if ((cl_text == adjust_start->get_type()) && (prev_end.get_next() != adjust_start->get_start()))
                    /* lines on next page are not synchronized */
                    /*   may happen when page format_ok is set or page is visible */
                    /*   in that case get_tused() does NOT reformat page */
                    break;
                }
            if (NULL == next_page)
                /* nothing remains that needs adjustment */
                adjust_start = NULL;
            }

        /* new line exceeds height of old line */
        /* resize "this" to size of new line and correct baseline */
        if (new_cline->get_hi() > get_hi()){
            for(start_line = this; start_line->left; start_line = start_line->left)
                ;
            /* only head tline of "this" must be resized as only it will be reused */
            start_line->set_baseline(new_tline->get_baseline());
            start_line->resize(new_cline->get_hi() - get_hi());
            for (start_line = left; start_line; start_line = start_line->left)
                start_line->set_baseline(new_tline->get_baseline());
            }
        /* old line exceeds height of new line */
        /* resize tline members of new line to size of "this" and correct baselines */
        else if (new_cline->get_hi() < get_hi())
            for (start_line = new_tline; start_line; start_line = start_line->right) {
                /* each tline of new_cline must be resized as all will be used */
                start_line->set_hi(get_hi());
                start_line->set_baseline(get_baseline());
                }

        /* "this" is invalid, so release its right queue members */
        if (right)
            right->release();
        /* reuse "this" to contain newly formatted material */
        right = new_tline->right;
        if (right) {
            right->left = this;
            right->set_prior(inq_prior());
            right->set_next(inq_next());
            }
        use_font = new_tline->use_font;
        start_pos = new_tline->start_pos;
        baseline_y = new_tline->baseline_y;
        text_width = new_tline->text_width;
        count = new_tline->count;
        start_x_offset = new_tline->start_x_offset;
        /* sets upper left and upper right of right siblings, too */
        set_ul_coord(get_ulx(), get_uly());

        /* indicate whether next line needs rerendering */
        /*   next line is OK if new line ends where old line ended */
        /*   otherwise remainder of viewport may need to be redrawn */
        if (0 == (next_ok = (old_end == new_end))) {
            /* antiflicker tests */
            if (old_end.in_node &&
              old_end.in_node->get_node(n_right) &&
              old_end.in_node->get_node(n_right)->get_node(n_right) &&
              new_end.in_node == old_end.in_node->get_node(n_right)->get_node(n_right)) {
                /* insertion has split the node in the end position of the old line */
                /* check if new end is actually the same character as old end */
                if (0 < (index_diff = old_end.char_index - ((gnode*) old_end.in_node)->get_char_count())) {
                    if (index_diff == new_end.char_index)
                        next_ok = 1;
                    }
                }
            else if (old_next.in_node && new_next.in_node &&
              (((char*) ((gnode*) old_next.in_node)->get_contents()) + old_next.char_index) ==
              (((char*) ((gnode*) new_next.in_node)->get_contents()) + new_next.char_index))
                /* deletion preserved line boundaries */
                next_ok = 1;
            }
        else if (old_next == new_next)
            /* deletion did not split node for beginning of next line */
            adjust_start = NULL;

        /* delete new head tline whose values reset "this" */
        new_tline->right = NULL;
        new_tline->release();
        /* set page head members of right siblings */
        for (hold_line = right; hold_line; hold_line = hold_line->right)
            hold_line->set_phead(get_phead());
        /* prepare to return reformatted line */
        return_line = this;
        }

    /* check for page overflow */
    depth = return_line->get_uly(); 
    for (overflow = return_line;
      overflow;
      overflow = overflow->inq_next()) {
        depth += overflow->get_hi();
        if (depth > overflow->get_phead()->get_hi())
            break;
        }
    /* on overflow: */
    /*   correct bottom_line of phead */
    /*   correct end_pos of phead and start_pos of next phead if any */
    /*   mark next page if any as needing reformatting */
    /*   release overflow */
    if (overflow) {
        overflow_y = overflow->get_uly();
        /* signal to incremental view_port::redraw() that view_pos */
        /*   objects associated with lines from overflow down are invalid */
        next_ok = 0;
        /* remember what new bottom line will be */
        new_bottom_line = overflow->inq_prior();
        /* remember page overflowed */
        old_page = new_bottom_line->get_phead();
        /* remember start position for next page */
        new_page_start = overflow->get_start();
        if (cl_text == overflow->get_type()) {
            if (overflow->get_start().char_index >= ((gnode*) overflow->get_start().in_node)->get_char_count()) {
                if (overflow->get_start().in_node &&
                  overflow->get_start().in_node->get_node(n_right) &&
                  overflow->get_start().in_node->get_node(n_right)->get_node(n_right)) {
                    /* insertion has split the node in the start position of the overflow line */
                    /* get model position of character referenced by overflow start_pos */
                    /*   before insertion */
                    new_page_start.in_node = overflow->get_start().in_node->get_node(n_right)->get_node(n_right);
                    new_page_start.char_index = overflow->get_start().char_index - ((gnode*) overflow->get_start().in_node)->get_char_count();
                    }
                }
            }
        /* reset bottom used line and end position of page */
        old_page->set_bottom_used(new_bottom_line);
        old_page->set_end_pos(new_page_start.get_prior());

        if (adjust_start && adjust_start->get_phead() == get_phead())
            page_after_overflow = get_phead()->get_next();
        else
            page_after_overflow = NULL;
        if (page_after_overflow) {
            page_after_overflow->set_start_pos(new_page_start);
            page_after_overflow->mark_reformat();
            }

        if (adjust_start
          && (adjust_start->get_uly() >= overflow_y)
          && (adjust_start->get_phead() == get_phead())) {
            /* adjust start falls within overflow from this page */
            if (NULL == get_phead()->get_next())
                adjust_start = NULL;
            else {
                for (next_page = page_after_overflow, next_page_start = new_page_start;
                  next_page;
                  next_page = next_page->get_next()) {
                    /* set start_pos of next page correctly in case call to phead::get_tused() */
                    /*   reformats page (will happen if page format is not OK and page is not visible) */
                    /* if page is reformatted, adjust_start will be moved to beginning of next page */
                    /*   by next iteration of loop, or will be set to NULL if there is no next page */
                    /* since adjust_start is moved beyond page, no need to check that start_pos of page */
                    /*   is not reset in adjustment loop at bottom of this function */
                    next_page->set_start_pos(next_page_start);
                    /* NOTE: get_tused() may cause next page to be reformatted */
                    for (hold_cline = next_page->get_tused();
                      hold_cline->inq_next() && (cl_text != hold_cline->get_type());
                      hold_cline = hold_cline->inq_next())
                        ;
                    adjust_start = hold_cline;
                    if ((cl_text == adjust_start->get_type()) && (next_page_start != adjust_start->get_start()))
                        /* lines on next page are not synchronized */
                        /*   may happen when page format_ok is set or page is visible */
                        /*   in that case get_tused() does NOT reformat page */
                        break;
                    /* NOTE: get_bused() will NOT cause next page to be reformatted as previous call */
                    /*   of next_page->get_tused() reformatted that page if necessary and possible */
                    for (last_line = next_page->get_bused(); 
                      last_line->inq_prior() && (cl_text != last_line->get_type()); 
                      last_line = last_line->inq_prior())
                        ;
                    if (cl_text == last_line->get_type())
                        next_page_start = ((tline*) last_line)->get_end().get_next();
                    }
                if (NULL == next_page)
                    /* nothing remains that needs adjustment */
                    adjust_start = NULL;
                }
            }

        /* free overflow lines */
        /* NOTE: must be done before update of bottom line */
        /*   because it sets page bottom_line to page bottom_free */
        overflow->get_phead()->free_clines(overflow);
        }

    /* underflow handled in incremental view_port::redraw() */

    /* correct any start or end positions affected by a node split if:                      */
    /*  (1) (a) return_line ends on an old line boundary and so following lines, if any,    */
    /*          will be retained                                                            */
    /*      AND                                                                             */
    /*      (b) at least one line remains as a known candidate for updating                 */
    /*  AND                                                                                 */
    /*  (2) (a) new line was not inserted before old                                        */
    /*      OR                                                                              */
    /*      (b) next position to be formatted differs from position following end of        */
    /*          old line but both have the same node identifier, indicating model_position  */
    /*          corrections are required                                                    */
    if (adjust_start) {
        /* correct start and end positions of text lines from adjust_start */
        /*   to end of material dominated by node in which old line ended */
        /* this material is the portion moved to a new right sibling node in a node split */
        if ((old_end == new_end) &&
          (old_next.in_node == new_next.in_node) &&
          (old_next.char_index != new_next.char_index)) {
            /* deletion was achieved by moving goo_ptr of a node */
            /*   whose first formatted character is in the reformatted line */
            /*   AND */
            /*   whose last formatted character appears somewhere in next line */
            /*     or below */
            old_node = hold_node = old_next.in_node;
            index_inc = old_next.char_index - new_next.char_index;
            }
        for (old_node = hold_node = NULL, prev_cline = adjust_start;
          prev_cline;
          prev_cline = next_cline) {
            if (prev_cline->get_type() == cl_text) {
                for (hold_line = (tline*) prev_cline;
                  hold_line;
                  hold_line = hold_line->right) {
                    if (NULL == hold_node) {
                        if (hold_line->start_pos.char_index >= ((gnode*) hold_line->start_pos.in_node)->get_char_count()) {
                            /* detect right sibling from split of node in which insertion */
                            /*   or deletion took place */
                            old_node = hold_node = hold_line->start_pos.in_node;
                            hold_node = hold_node->get_right();
                            if (NULL == hold_node)
                                break;
                            index_inc = old_next.char_index - new_next.char_index;
                            if (hold_node != new_next.in_node) {
                                /* insertion */
                                hold_node = hold_node->get_right();
                                if (NULL == hold_node)
                                    break;
                                index_inc = ((gnode*) old_node)->get_char_count();
                                }
                            }
                        }
                    if (old_node == hold_line->start_pos.in_node) {
                        hold_line->start_pos.in_node = hold_node;
                        hold_line->start_pos.char_index -= index_inc;
                        }
                    }
                }
            if (NULL == (next_cline = prev_cline->inq_next())) {
                /* have reached end of page */
                /* set start position ... */
                /* NOTE: get_tused() may cause page to be reformatted */
                for (hold_cline = prev_cline->get_phead()->get_tused();
                  hold_cline->inq_next() && (cl_text != hold_cline->get_type());
                  hold_cline = hold_cline->inq_next())
                    ;
                if ((cl_text == hold_cline->get_type()) && (page_after_overflow != hold_cline->get_phead()))
                    hold_cline->get_phead()->set_start_pos(((tline*) hold_cline)->start_pos);
                /* ... and end position of page, checking that start position of */
                /*   page following overflowed page is not reset */
                /* NOTE: get_bused() will NOT cause page to be reformatted as previous call */
                /*   to get_tused() reformatted that page if necessary and possible */
                for (hold_cline = prev_cline->get_phead()->get_bused();
                  hold_cline->inq_prior() && (cl_text != hold_cline->get_type());
                  hold_cline = hold_cline->inq_prior())
                    ;
                if (cl_text == hold_cline->get_type()) {
                    for (hold_line = (tline*) hold_cline;
                      hold_line->right;
                      hold_line = hold_line->right)
                        ;
                    hold_line->get_phead()->set_end_pos(hold_line->get_end());
                    }
                if ((NULL != hold_cline) && (NULL != hold_cline->get_phead()->get_next())) {
                    /* NOTE: get_tused() may cause next page to be reformatted */
                    /* correct start position of next page before that call */
                    hold_cline->get_phead()->get_next()->set_start_pos(hold_cline->get_phead()->get_end().get_next());
                    for (next_page = hold_cline->get_phead()->get_next(),
                      next_page->set_start_pos(hold_cline->get_end().get_next()),
                      next_cline = hold_cline->get_phead()->get_next()->get_tused();
                      next_cline && next_page;
                      next_page = next_page->get_next()) {
                        for (;
                          next_cline->inq_next() && (cl_text != next_cline->get_type()); 
                          next_cline = next_cline->inq_next())
                            ;
                        if (next_cline->get_start() == hold_cline->get_end().get_next()) {
                            /* lines on next page are synchronized */
                            if (NULL == next_page->get_next())
                                /* no more pages, no more lines */
                                next_cline = NULL;
                            else {
                                for (prev_cline = next_page->get_bused();
                                  prev_cline->inq_prior() && (cl_text != prev_cline->get_type());
                                  prev_cline = prev_cline->inq_prior())
                                    ;
                                /* NOTE: get_down() may cause next page to be reformatted */
                                /* correct start position of next page before that call */
                                if ((cl_text == prev_cline->get_type()) && (next_page->get_next()))
                                    next_page->get_next()->set_start_pos(prev_cline->get_end().get_next());
                                next_cline = prev_cline->get_down();
                                }
                            }
                        else
                            break;
                        }
                    }
                }
            }
        }
          
    return return_line;
}

/* draw a text line--includes drawing any of its right members */
/* argument start_index specifies an offset into the string dominated by the line */
/* it is used by model_pos::in_place_redraw() to specify the offset within the line */
/*   of the character dominated by the model_pos */
view_pos* tline::draw_line(view_port *in_viewport, screen_xy where, long start_index)
{
    register view_pos  *return_pos;
    register char      *do_draw_st;
    register tline     *do_line;
             vpftbl    *cur_ftbl;
             char_set  *do_char_set;
    unsigned long       do_char_size,
                        this_draw_len, do_draw_len,
                        do_lead_tabs, do_trail_tabs;
             long      *do_inv_prop_val;
             long       this_start_index, do_width;
             char      *this_draw_st;
             char      *last_char;
             node      *do_node;
             model_pos  lcur_pos, lleft_pos;
             screen_xy  do_where, do_adj_where;
             lrud_dir   do_text_dir;
             long      *format_value;
             
    if (cl_text != get_type())
        /* handles tline of type cl_gutter containing formfeed */
        /*   when formfeed property enabled */
        return cline::draw_line(in_viewport, where, start_index);

    return_pos = in_viewport->get_pos();
    if (!return_pos) return 0;
    return_pos->set_x(where.at_x);
    return_pos->set_y(where.at_y);
    if ((start_index < 0) || (start_index < start_pos.char_index))
        start_index = 0;
    else
        start_index = start_index - start_pos.char_index;

    do_where = where;

    cur_ftbl = in_viewport->get_ftbl();

    /* sentinel determines whether anything drawn by first for loop */
    this_draw_len = 0;
    /* sentinel marks where midline redraw began */
    this_start_index = start_index;

    /* draw queue of tlines beginning with this tline */
    for (do_line = this, do_lead_tabs = 0, do_trail_tabs = 0;
      do_line; 
      do_line = do_line->right, do_lead_tabs = 0, do_trail_tabs = 0) {
        if (0 == (do_node = do_line->start_pos.in_node))
            /* no text in do_line */
            continue;
        do_inv_prop_val = (long *) in_viewport->inq_state(do_node, tml_inv_prop);
        do_char_set = (char_set*) do_node->get_model()->inq_state(do_node, tml_char_set_prop);
        do_char_size = do_char_set->get_char_size();
        do_draw_st = ((char*) do_node->get_contents()) + do_line->start_pos.char_index + start_index;
        do_draw_len = do_line->count - start_index;
        last_char = do_draw_st + do_draw_len - do_char_size;
        /* don't draw newline at EOL */
        if (do_char_set->is_newline((unsigned char*) last_char))
            do_draw_len -= do_char_size;
        /* don't draw trailing tabs (in segment returned by gnode::tamrof_line()) */
        while (do_draw_len && do_char_set->is_htab((unsigned char*) last_char)) {
            ++do_trail_tabs;
            do_draw_len -= do_char_size;
            last_char -= do_char_size;
            }
        /* don't draw leading tabs (in segment returned by gnode::format_line()) */
        if (do_char_set->is_htab((unsigned char*) do_draw_st))
            do_lead_tabs = 1;
        while (do_draw_len && do_char_set->is_htab((unsigned char*) do_draw_st)) {
            do_draw_len -= do_char_size;
            do_draw_st += do_char_size;
            }
        /* don't draw leading space when word_wrap enabled */
        do_text_dir = do_char_set->get_dir_to_char((unsigned char*)do_draw_st);
        if ((! do_lead_tabs)
          && (format_value = (long*) get_format()->inq_state(start_pos.in_node, tml_wrap_prop))
          && (word_wrap == *format_value)
          && (((::right == do_text_dir) && (0 == get_ulx()))
             || 
             ((::left == do_text_dir) && ((get_ulx() + get_wide()) == get_phead()->get_wide()))
             || 
             (((::down == do_text_dir) /*|| (::stack_down == do_text_dir)*/) && (0 == get_uly()))
             ||
             (((::up == do_text_dir)   /*|| (::stack_up == do_text_dir)*/) && ((get_uly() + get_hi()) == get_phead()->get_hi())))) {
            while (do_draw_len && do_char_set->cs_isspace((unsigned char*) do_draw_st)) {
                do_draw_len -= do_char_size;
                do_draw_st += do_char_size;
                }
            }

        if ((do_lead_tabs) || (!start_index))
            /* correct starting position for text draw by tab offset and margin at beginning of line */
            /* when leading tabs not present, correct for margin at beginning of line */
            do_where.at_x = do_line->get_ulx() + do_line->get_start_x_off() + in_viewport->get_hoffset();

        if (0 == this_draw_len) {
            /* preserve values for later use in determining overlap */
            /* use test for this_draw_len to skip lines with no contents */
            this_draw_st = do_draw_st;
            this_draw_len = do_draw_len;
            }

        if (do_draw_len) {

            /* have something to draw in do_line */

            /* Note that when start_index is not zero only part of this is cleared and   */
            /*   redrawn, requiring check for overlap with last character not redrawn.   */
            /* TMLib protocol for drawing lines:                                         */
            /* (1) Clip to bounding box of tline-and-right-siblings.                     */
            /* (2) Clear bounding box of tline-and-right-siblings.                       */
            /* (3) Clear clip.                                                           */
            /* (4) Draw tline and each of its right siblings.                            */
            /* (5) Redraw tlines to left of this, starting with first tline with origin  */
            /*     left of overlap region.                                               */
            /* (6) When start_index is within line, redraw overlapped characters to left.*/
            /* Steps (1) through (3) are performed before this function is called.       */
            /* Steps (4) through (6) follow immediately.                                 */
            /* In graphics systems that report expose events, the clearing is performed  */
            /*   once for the entire exposure in tml_c_repaint_screen() (called by       */
            /*   tml_dp_damage_repair).                                                  */
            /* This protocol ensures that any overlapping portions of lines are properly */
            /*   restored. Lines may overlap when e.g. italic fonts, fancy script fonts, */
            /*   or Arabic fonts are used.                                               */
            /* The redrawing strategy works because tml_dp_string_draw() renders the     */
            /*   text in foreground color against a transparent background.              */

            /* get baseline origin for text drawing call */
            do_adj_where = do_where;
            do_adj_where.at_y += do_line->baseline_y;

            /* correct x baseline origin for right-to-left text to right end of box */
            if (::left == do_text_dir) {
                do_width = (*cur_ftbl->ptr_2_tml_dp_string_size)(do_draw_st, do_draw_len,
                                                                 do_line->use_font->get_handle());
                do_adj_where.at_x += do_width;
                }

            /* draw the string for do_line */
            do_width = (*cur_ftbl->ptr_2_tml_dp_string_draw)(in_viewport->get_handle(),
                                                             do_draw_st, do_draw_len,
                                                             do_adj_where.at_x, do_adj_where.at_y,
                                                             do_line->use_font->get_handle());

            /* invert the string for do_line if invert property enabled */
            if (do_inv_prop_val && *do_inv_prop_val)
                (*cur_ftbl->ptr_2_tml_dp_blt)(in_viewport->get_handle(),
                                              do_where.at_x, do_where.at_y,
                                              do_where.at_x + do_width,
                                              do_where.at_y + do_line->get_hi(),
                                              do_where.at_x, do_where.at_y,
                                              TML_DP_INV);
            
            /* increment x origin by width of do_line */
            do_where.at_x += do_width;
            if (do_trail_tabs)
                do_where.at_x += calc_tab_pixel_off(do_line->get_start().in_node,
                                                    do_line->get_format(),
                                                    do_trail_tabs,
                                                    do_where.at_x);
            }
        /* start subsequent lines at their beginning */
        start_index = 0;
        }

    /* repaint any overlapped region of lines to left */
    if (left)
        restore_left_lines(in_viewport, where, this_draw_st, this_draw_len);

    /* repaint any overlapped portion of leftward portion of this line */
    if (this_start_index)
        restore_left_fragment(in_viewport, where, this_draw_st, this_start_index);

    /* increment x member of origin argument by total width drawn */
    where = do_where;

    /* associate redrawn lines with view position returned */
    return_pos->set_line(this);
    return return_pos;
}

void tline::restore_left_lines(view_port* in_viewport, screen_xy where, char* this_draw_st, unsigned long this_draw_len)
{
             char_set  *this_char_set, *lcur_char_set;
    unsigned long       this_char_size, lchar_size, ldraw_len, lhave_tabs, lchar_val;
             long      *linv_prop_val, *this_inv_prop_val;
             long       i, left_corr, lwidth, charp_incr;
             char      *lcur_char, *ldraw_st, *this_char;
             node*      lcur_node;
             tline*     lcur_line;
             vpftbl*    cur_ftbl;
             model_pos  lleft_pos, lcur_pos;
             xy_coord   lcur_xy_coord;
             screen_xy  lwhere, ladj_where,lcur_screen_xy, corr_box_start, corr_box_end;
             lrud_dir   this_text_dir, ltext_dir;

             typedef model_pos (model_pos::*mp_iter_func_p)();
             typedef long      (model_pos::*mp_comp_func_p)(model_pos);

             mp_iter_func_p   iter_func;
             mp_comp_func_p   comp_func;

    /* there is a line to the left to be overlapped */

    this_inv_prop_val = (long*) start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_inv_prop);
    this_char_set = (char_set*) start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_char_set_prop);
    this_char_size = this_char_set->get_char_size();
    this_text_dir = this_char_set->get_dir_to_char((unsigned char*)this_draw_st);
    cur_ftbl = in_viewport->get_ftbl();

    /* establish initial coordinates for correction box */
    corr_box_start = corr_box_end = where;

    if (0 != this_draw_len) {
        /* something was drawn by caller tline::draw_line */
        /* get overlap due to leftmost nonspace character in this */
        /*   and its right siblings. */
        /* first character in string is leftmost when text runs left */
        /* last character in string is leftmost when text runs right */
        if (::right == this_text_dir) {
            this_char = this_draw_st;
            charp_incr = this_char_size;
            }
        else if (::left == this_text_dir) {
            this_char = this_draw_st + this_draw_len - this_char_size;
            charp_incr = -(this_char_size);
            }
        for (i = this_draw_len;
          i && (this_char_set->cs_isspace((unsigned char*)this_char));
          i -= this_char_size)
            this_char += charp_incr;
        
        if (i) {
            /* lbearing of leftmost nonspace character is maximum possible overlap */
            left_corr = use_font->get_lbearing(this_char_set->as_ulong((unsigned char*)this_char));
            if (::left == this_text_dir) {
                /* text runs left; correct for width of character and */
                /*   intercharacter space because they are within box */
                left_corr -= use_font->get_width(this_char_set->as_ulong((unsigned char*)this_char));
                left_corr -= use_font->get_icspace();
                }
            corr_box_start.at_x += left_corr;
            }
        }

    /* even if nothing was drawn by caller tline::draw_line, must check for overlap */
    /*   by lines to left of this into space cleared before that call */
    /* get overlap due to rightmost nonspace character in lines to left of this */
    /* search lines to left for rightmost nonspace character */
    for (lcur_line = left; lcur_line; lcur_line = lcur_line->left) {
        if (0 == (lcur_node = lcur_line->start_pos.in_node))
        /* no text in line */
            continue;

        lcur_char_set = (char_set*) lcur_node->get_model()->inq_state(lcur_node, tml_char_set_prop);
        lchar_size = lcur_char_set->get_char_size();
        ldraw_st = ((char*) lcur_node->get_contents()) + lcur_line->start_pos.char_index;

        lleft_pos = lcur_pos = lcur_line->start_pos;

        /* Find leftmost and rightmost model position in line and choose */
        /*   iteration and comparison functions for loop below. */
        /* The iteration functions move from rightmost to leftmost character */
        /*   via model position moves. They return bad model positions when */
        /*   no move is possible, thereby handling model boundaries. */
        /* The comparison functions tell whether the leftmost character has been */
        /*   exceeded by the last move. They return false when either of their */
        /*   operands is a bad model position. */

        if (::right == (ltext_dir = lcur_char_set->get_dir_to_char((unsigned char*)ldraw_st))) {
            /* rightmost character is last in string when text runs right */
            lcur_pos.char_index += (lcur_line->count - lchar_size);
            iter_func = &model_pos::get_prior;
            comp_func = &model_pos::operator>=;
            }
        else if (::left == ltext_dir) {
            /* leftmost character is last in string when text runs left */
            lleft_pos.char_index += (lcur_line->count - lchar_size);
            iter_func = &model_pos::get_next;
            comp_func = &model_pos::operator<=;
            }

        /* loop from rightmost to leftmost character in line */
        /*   seeking nonspace character */
        for (;
          (lcur_pos.*comp_func)(lleft_pos);
          lcur_pos = (lcur_pos.*iter_func)()) {
            lcur_char = ((char*) lcur_node->get_contents()) + lcur_pos.char_index;
            if (0 == lcur_char_set->cs_isspace((unsigned char*)lcur_char))
                /* nonspace character found */
                break;
            }

        /* rightmost line containing nonspace character found */
        if ((lcur_pos.*comp_func)(lleft_pos))
            break;
        }

    if (0 != lcur_line) {
        /* nonspace character found to left */
        /* get origin xy position on conceptual page of character found */
        lcur_xy_coord  = lcur_line->mpos_2_xy(lcur_pos);
        lcur_screen_xy.at_x = lcur_xy_coord.x_coord;

        /* correct for viewport's horizontal offset with respect to conceptual page */
        /*   and for farthest rightward extent of character */
        lchar_val = lcur_char_set->as_ulong((unsigned char*)lcur_char);
        lcur_screen_xy.at_x += in_viewport->get_hoffset()
                               + lcur_line->use_font->get_rbearing(lchar_val);

        if (lcur_screen_xy.at_x > where.at_x)
            corr_box_end.at_x = lcur_screen_xy.at_x;

        /* seek rightmost of lines to left that begin left of correction box */
        while (lcur_line &&
          (corr_box_start.at_x < (lcur_line->get_ulx() + in_viewport->get_hoffset())))
            lcur_line = lcur_line->left;

        /* correct y members of correction box corners */
        corr_box_start.at_y += lcur_line->baseline_y;
        corr_box_end.at_y += lcur_line->baseline_y;

        /* paint in the overlapped lines */
        for (lhave_tabs = 0;
          lcur_line && (lcur_line != this);
          lcur_line = lcur_line->right, lhave_tabs = 0) {
            if (0 == (lcur_node = lcur_line->start_pos.in_node))
                /* no text in this line */
                continue;
            lcur_char_set = (char_set*) lcur_node->get_model()->inq_state(lcur_node, tml_char_set_prop);
            lchar_size = lcur_char_set->get_char_size();
            ldraw_st = ((char*) lcur_node->get_contents()) + lcur_line->start_pos.char_index;
            linv_prop_val = (long *) in_viewport->inq_state(lcur_line->start_pos.in_node, tml_inv_prop);
            ldraw_len = lcur_line->count;
            if (lcur_char_set->is_htab((unsigned char*)ldraw_st))
                lhave_tabs = 1;
            while(lcur_char_set->is_htab((unsigned char*)ldraw_st)) {
                ldraw_len -= lchar_size;
                ldraw_st += lchar_size;
                }
            lwhere = in_viewport->mpos_2_xy(lcur_line->start_pos);
            if (lhave_tabs)
                lwhere.at_x = lcur_line->get_ulx() + lcur_line->get_start_x_off() + in_viewport->get_hoffset();

            if (ldraw_len) {
                lwhere.at_y = where.at_y;
                ladj_where = lwhere;
                ladj_where.at_y += lcur_line->baseline_y;

                /* redraw string for overlapped line to left */
                if (::left == (ltext_dir = lcur_char_set->get_dir_to_char((unsigned char*)ldraw_st))) {
                    /* text runs left; adjust drawing origin to rightmost character origin */
                    lwidth = (*cur_ftbl->ptr_2_tml_dp_string_size)(ldraw_st, ldraw_len,
                                                                   lcur_line->use_font->get_handle());
                    ladj_where.at_x += lwidth;
                    }

                lwidth = (*cur_ftbl->ptr_2_tml_dp_string_draw)(in_viewport->get_handle(),
                                                               ldraw_st, ldraw_len,
                                                               ladj_where.at_x, ladj_where.at_y,
                                                               lcur_line->use_font->get_handle());

                /* invert the string for overlapped line to left if invert property enabled */
                /* avoid reinverting previously inverted material */
                if (this_inv_prop_val && *this_inv_prop_val && linv_prop_val && *linv_prop_val) {
                    if (where.at_x < lwhere.at_x + lwidth)
                        lwidth -= (lwhere.at_x + lwidth - where.at_x);
                    (*cur_ftbl->ptr_2_tml_dp_blt)(in_viewport->get_handle(),
                                                  lwhere.at_x, lwhere.at_y,
                                                  lwhere.at_x + lwidth,
                                                  lwhere.at_y + lcur_line->get_hi(),
                                                  lwhere.at_x, lwhere.at_y,
                                                  TML_DP_INV);
                    }
                }
            }
        }
}

void tline::restore_left_fragment(view_port* in_viewport, screen_xy where, char* this_draw_st, long this_start_index)
{
             char      *this_char, *lcur_char, *ldraw_st;
             char_set  *this_char_set;
    unsigned long       this_char_size, lcur_char_val;
    lrud_dir            this_text_dir;
             long      *this_inv_prop_val;
             long       i, left_corr;
             vpftbl    *cur_ftbl;
             model_pos  lleft_pos, max_left;
             screen_xy  corr_box_start, lcur_screen_xy;

             typedef model_pos (model_pos::*mp_iter_func_p)();

             mp_iter_func_p   left_func;
             mp_iter_func_p   right_func;

    this_char = this_draw_st;
    this_char_set = (char_set*) start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_char_set_prop);
    this_char_size = this_char_set->get_char_size();
    this_text_dir = this_char_set->get_dir_to_char((unsigned char*)this_char);
    this_inv_prop_val = (long*) start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_inv_prop);
    cur_ftbl = in_viewport->get_ftbl();

    /* midline redraw */
    /* redraw initial portion of line if required */

    /* correct start x origin of correction box */
    /*   by left overlap from first character redrawn */
    corr_box_start = where;
    left_corr = use_font->get_lbearing(this_char_set->as_ulong((unsigned char*)this_char));
    if (::left == this_text_dir) {
        /* text runs left; correct for width of character and */
        /*   intercharacter space because they are within box */
        left_corr -= use_font->get_width(this_char_set->as_ulong((unsigned char*)this_char));
        left_corr -= use_font->get_icspace();
        }
    corr_box_start.at_x += left_corr;

    /* count characters from beginning of this that are not overlapped */

    /* find leftmost position in portion of line not redrawn */
    lleft_pos = start_pos + this_start_index;
    if (::right == this_text_dir) {
        left_func = &model_pos::get_prior;
        right_func = &model_pos::get_next;
        max_left = start_pos;
        }
    else if (::left == this_text_dir) {
        left_func = &model_pos::get_next;
        right_func = &model_pos::get_prior;
        max_left = start_pos += count - this_char_size;
        }
    lleft_pos = (lleft_pos.*left_func)();
    if (this_char_set->is_htab(((unsigned char*) lleft_pos.in_node->get_contents()) + lleft_pos.char_index))
        return;

    /* find leftmost nontab character */
    for (;
      this_char_set->is_htab(((unsigned char*) max_left.in_node->get_contents()) + max_left.char_index);
      max_left = (max_left.*right_func)(), this_start_index -= this_char_size)
        ;

    /* find leftmost character that is overlapped */
    for (i = 0; i < this_start_index; i += this_char_size) {
        lcur_screen_xy = in_viewport->mpos_2_xy(lleft_pos);
        lcur_char = ((char*) lleft_pos.in_node->get_contents()) + lleft_pos.char_index;
        lcur_char_val = this_char_set->as_ulong((unsigned char*) lcur_char);
        if (corr_box_start.at_x > (lcur_screen_xy.at_x + use_font->get_rbearing(lcur_char_val))) {
            /* rightmost nonoverlapped character found */
            lleft_pos = (lleft_pos.*right_func)();
            break;
            }
        if (lleft_pos == max_left) {
            /* have reached first character in line */
            i += this_char_size;
            break;
            }
        lleft_pos = (lleft_pos.*left_func)();
        }

    if (i) {

        /* overlap exists */

        /* get characters to draw */
        lcur_screen_xy = in_viewport->mpos_2_xy(lleft_pos);
        ldraw_st = ((char*) lleft_pos.in_node->get_contents()) + lleft_pos.char_index;

        /* redraw overlapped characters */
        (*cur_ftbl->ptr_2_tml_dp_string_draw)(in_viewport->get_handle(), 
                                              ldraw_st, i,
                                              lcur_screen_xy.at_x, lcur_screen_xy.at_y,
                                              use_font->get_handle());
        }
}

