/**************************************************************************************
    csm_kmp.hxx
    header file for the object definition used to implement Knuth-Morris-Pratt
        string-search algorithm (KMP) -- O(m+n) and no backtracking

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef csm_kmp_hxx
#define csm_kmp_hxx

#include <stdio.h>

#ifndef csm_character_hxx
#include <tml/csm_character.hxx>
#endif

#ifndef char_set_mgr_hxx
#include <tml/char_set_mgr.hxx>
#endif

class kmp_node;
class kmp_re_node;
class kmp_equiv_node;
class kmp_class_node;
class kmp_range_node;
class kmp_list_node;
class kmp_newline_node;

/**********************************************************************************
    Objects used to implement Knuth-Morris-Pratt algorithm
**********************************************************************************/

/* enumerated class of navigation directions for a kmp_node */
enum kmp_dir { kmp_up, kmp_right, kmp_down };

class csm_string {
public:
    csm_character* string;
    unsigned long  len;
};

enum matches_kmp_node { yes = -1, no, maybe, unknown, defined };

class kmp_node {
protected:
    kmp_node*      up;
    kmp_node*      right;
    kmp_node*      down;
    csm_character* this_character;
    kmp_node*      failure;
    unsigned char* matched;

    friend kmp_node* kmp_fsm(unsigned char**, unsigned long, unsigned long*, char_set* = 0);
public:
    kmp_node();        
    ~kmp_node();       
    virtual kmp_node*     next_live_parent(){return this;}
    virtual kmp_node*     get_up(){return (up ? up->next_live_parent() : up);}
    virtual kmp_node*     next_live_child(){return this;}
    virtual kmp_node*     get_down(){return (down ? down->next_live_child() : down);}
    virtual kmp_node*     next_live_sibling(){return this;}
    virtual kmp_node*     get_right(){return (right ? right->next_live_sibling() : right);}
    virtual kmp_node*     get_failure(){return failure;}
    csm_character*        get_character(){return this_character;}
    unsigned char*        get_matched(){return matched;}
    void                  set_up(kmp_node* new_up){up = new_up;}
    void                  set_right(kmp_node* new_right){right = new_right;}
    void                  set_down(kmp_node* new_down){down = new_down;}
    void                  set_failure(kmp_node* new_failure){failure = new_failure;}
    void                  set_character(csm_character* new_character){this_character = new_character;}
    void                  set_matched(unsigned char* new_match){matched = new_match;}
    kmp_node*             breadth_first(kmp_dir& dir, unsigned long& levels);
    kmp_node*             depth_first(kmp_dir& dir);

    // Match a particular input character against a given kmp_node "this" which also records a character
    virtual  long         matches(unsigned char* s){return this_character->eq(s, (unsigned long) 1);}
    virtual  long         matches(kmp_node*       other_node){return matches(other_node->this_character->get_value());}

    virtual kmp_node* check_lookup_backup(kmp_node* cur_node, unsigned char* s, unsigned long char_size);
    virtual kmp_node* check_incr_lookup_backup(kmp_node* next_node, unsigned char* s, unsigned long char_size);

    unsigned long lookup(unsigned char*& sought);
    unsigned long incr_lookup(kmp_node*& next_node, unsigned long& num_chars, unsigned char* s);
};

#endif
// csm_kmp_hxx
