/*************************************************************************************
    csm_lock.cxx
    member functions of objects used for lock management in multiprocess applications

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef csm_lock_hxx
#include <tml/csm_lock.hxx>
#endif

#include <stdio.h>
#include <tml/tml_os_depends.hxx>

/*************************************************************************************
    GLOBAL PER-PROCESS PROCESS IDENTIFIER FOR SELF
*************************************************************************************/

// To make this global per-process, must be declared with attribute specifier
//   for Apollo builds to allow use in global and inlib'd libraries
#ifdef apollo
void* csm_my_pid __attribute((section(csm_my_pid))) = NULL;
#else
void* csm_my_pid = NULL;
#endif

/*************************************************************************************
    INTERFACE TO GET OWN PID
*************************************************************************************/

void* csm_get_my_pid()
{
    if (NULL == csm_my_pid)
        csm_my_pid = tml_os_getpid();
    return csm_my_pid;
}

/*************************************************************************************
    MEMBER FUNCTIONS FOR CLASS csm_proc_q
*************************************************************************************/

/* create a process queue element */
csm_proc_q::csm_proc_q(void* new_pid)
{
    pid = new_pid;
    next = NULL;
}

/* destroy a process queue element */
csm_proc_q::~csm_proc_q()
{
    next = NULL;
}

/*************************************************************************************
    MEMBER FUNCTIONS FOR CLASS csm_proc_q
*************************************************************************************/

/* create a lock for a unique resource */
csm_lock::csm_lock()
{
    tml_sem = tml_os_init_sem(1);
    head = NULL;
    tail = NULL;
    lockee = NULL;
}

/* create a lock for a specified unique resource */
csm_lock::csm_lock(void* new_lockee)
{
    tml_sem = tml_os_init_sem(1);
    head = NULL;
    tail = NULL;
    lockee = new_lockee;
}

/* create a lock for a resource pool */
csm_lock::csm_lock(long pool_capacity)
{
    tml_sem = tml_os_init_sem(pool_capacity);
    head = NULL;
    tail = NULL;
    lockee = NULL;
}

/* create a lock for a specified resource pool */
csm_lock::csm_lock(long pool_capacity, void* new_lockee)
{
    tml_sem = tml_os_init_sem(pool_capacity);
    head = NULL;
    tail = NULL;
    lockee = new_lockee;
}

/* destroy a csm_lock, killing all processes DEAD that sleep on it */
csm_lock::~csm_lock()
{
    csm_proc_q* cur_proc;

    tml_os_free_sem(tml_sem);
    tml_sem = NULL;
    for (cur_proc = head; cur_proc; cur_proc = cur_proc->next)
        tml_os_kill_proc(cur_proc->pid);
    while (head) {
        cur_proc = head;
        head = cur_proc->next;
        delete cur_proc;
        }
    head = NULL;
    tail = NULL;
    lockee = NULL;
}

/* busy wait on a lock */
void csm_lock::p_busy()
{
    while (tml_os_busy_wait(tml_sem))
        ;
    tml_os_p_sem(tml_sem);
}

/* wait on lock's queue if resource unavailable */
void csm_lock::p_suspend()
{
    tml_os_p_sem(tml_sem);
}

/* lock resource if available, returning 0, else return -1 */
long csm_lock::p_bounce()
{
    // the resource is available
    if (tml_os_resource_available(tml_sem)) {
        tml_os_p_sem(tml_sem);
        return 0;
        }
    else
        return -1;
}

/* unlock a resource */
void csm_lock::v()
{
    void*       next_pid;
    csm_proc_q* washead;

    if (NULL != head)
        next_pid = head->pid;
    else
        // always this case in sys5
        next_pid = NULL;

    if (tml_os_v_sem(next_pid, tml_sem)) {
        if (next_pid)
            fprintf(stderr, "csm_lock::v(): Failed to wake process %d\n", head->pid);
        else
            fprintf(stderr, "csm_lock::v(): Failed to wake an existing process\n");
        }
    else if (head) {
        washead = head;
        head = head->next;
        delete washead;
        }

}

