/**************************************************************************************
    model_pos_traits.cxx
    basic traits for the model_pos object

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>
#include <tml/tml_dp_graph.hxx>

/*basic constructor for a model position*/
model_pos::model_pos()
{
    in_node = 0;
    char_index = -1;     //i.e., not on a character
}

/**************************************************************************************
    logical comparisons--forms an algebra of positions in the model
**************************************************************************************/

/* correct model position to canonical form */
/*   i.e. with char_index that falls within scope of the node member */
void model_pos::normalize()
{
    long i;
    model_pos hold_pos;

    if ((0 == in_node)
      /* invalid node member */
      || (gobj != in_node->get_type())
      /* not normalizable since not a goo node */
      || ((char_index >= 0) && (char_index < ((gnode*) in_node)->get_char_count())))
      /* already normal */
        return;
    i = char_index;
    char_index = 0;
    hold_pos = *this + i;
    if (i && (hold_pos == *this)) {
        /* no goo at that distance */
        in_node = 0;
        char_index = -1;
        }
    else
        *this = hold_pos;
    return;
}

/*equivalence operator--do the two have the same contents?*/
long model_pos::operator== (model_pos comp_pos)
{
    if ((in_node == comp_pos.in_node) && (char_index == comp_pos.char_index))
        return 1;
    return 0;
}

/*non_eqivalence*/
long model_pos::operator!= (model_pos comp_pos)
{
    if ((in_node == comp_pos.in_node) && (char_index == comp_pos.char_index))
        return 0;
    return 1;
}

/*the less-than operator*/
long model_pos::operator< (model_pos comp_pos)
{
    register node *tn1,
                  *tn2;
    register long  tn1_level,
                   tn2_level;

    /*make sure both model positions are real; if not return false*/
    if ( !in_node || !comp_pos.in_node)
        return 0;

    if (in_node == comp_pos.in_node)
        return (char_index < comp_pos.char_index);

    /*they aren't the same node, so we need to find out if they are in the same model,
        and if so, which comes before the other*/
    tn1 = in_node;
    tn2 = comp_pos.in_node;
    if (tn1->get_model() != tn2->get_model())    //they aren't in the same model, so return false
        return 0;

    tn1_level = tn1->get_level();
    tn2_level = tn2->get_level();
                     
    /*get tn1 up to the same level as tn2*/
    for ( ; tn1_level > tn2_level; tn1 = tn1->get_node(n_up), tn1_level--)
        ;
    if (tn1 == tn2) return 0;   //tn2 is a parent of tn1; hence tn1 > tn2
                                                                                                 
    /*get tn2 up to the same level at tn1*/
    for ( ; tn2_level > tn1_level; tn2 = tn2->get_node(n_up), tn2_level--)
        ;
    if (tn2 == tn1) return 1;   //tn1 is a parent of tn2; hence tn1 < tn2
                                                                                                 
    /*get them to the point of having the same parent*/
    while (tn1->get_node(n_up) != tn2->get_node(n_up)){
        tn1 = tn1->get_node(n_up);
        tn2 = tn2->get_node(n_up);
        }
                                                                            
    /*compare their sibling order members*/
    if (tn1->get_sib_order() >= tn2->get_sib_order())
        return 0;
    return 1;
}
                                     
/*the less-than-or-equals operator*/
long model_pos::operator<= (model_pos comp_pos)
{
    if ((comp_pos == *this) || (*this < comp_pos))
        return 1;
    return 0;
}
                                                    
/*the greater-than operator*/
long model_pos::operator> (model_pos comp_pos)
{
    register node *tn1,
                  *tn2;
    register long  tn1_level,
                   tn2_level;

    /*make sure both model positions are real; if not return false*/
    if ( !in_node || !comp_pos.in_node)
        return 0;

    if (in_node == comp_pos.in_node)
        return (char_index > comp_pos.char_index);

    /*they aren't the same node, so we need to find out if they are in the same model,
        and if so, which comes before the other*/
    tn1 = in_node;
    tn2 = comp_pos.in_node;
    if (tn1->get_model() != tn2->get_model())    //they aren't in the same model, so return false
        return 0;

    tn1_level = tn1->get_level();
    tn2_level = tn2->get_level();
                     
    /*get tn1 up to the same level as tn2*/
    for ( ; tn1_level > tn2_level; tn1 = tn1->get_node(n_up), tn1_level--)
        ;
    if (tn1 == tn2) return 1;   //tn2 is a parent of tn1; hence tn1 > tn2
                                                                                                 
    /*get tn2 up to the same level at tn1*/
    for ( ; tn2_level > tn1_level; tn2 = tn2->get_node(n_up), tn2_level--)
        ;
    if (tn2 == tn1) return 0;   //tn1 is a parent of tn2; hence tn1 < tn2
                                                                                                 
    /*get them to the point of having the same parent*/
    while (tn1->get_node(n_up) != tn2->get_node(n_up)){
        tn1 = tn1->get_node(n_up);
        tn2 = tn2->get_node(n_up);
        }
                                                                            
    /*compare their sibling order members*/
    if (tn1->get_sib_order() <= tn2->get_sib_order())
        return 0;
    return 1;
}

/*the greater-than-or-equals operator*/
long model_pos::operator>= (model_pos comp_pos)
{
    if ((comp_pos == *this) || (*this > comp_pos))
        return 1;
    return 0;
}
                

/*the + operation for nonnegative integer increments */
model_pos model_pos::operator+ (long inc)
{
             model_pos return_pos;
             model_pos hold_pos;
    register long      i;

    /*check for trash*/
    if (!in_node)
        return return_pos;
          
    if (inc < 0)
        return *this - (-inc);

    return_pos = *this;

    if (0 == inc)
        return return_pos;

    i = in_node->get_max_index();
    i -= char_index;

    if (i >= inc) {
        return_pos.char_index += inc;
        return return_pos;
        }

    return_pos.char_index += i;
    return_pos = return_pos.get_next();
    if (!return_pos.in_node)
        return *this;
    inc -= (i + 1);

    if (0 == inc)
        return return_pos;

    hold_pos = return_pos;
    return_pos += inc;
    if (hold_pos == return_pos)
        return *this;
    else
        return return_pos;
}
                         
/*the - operator for nonnegative integer argument*/
model_pos model_pos::operator- (long dec)
{
             model_pos return_pos;
             model_pos hold_pos;
    register long      i;

    /*check for trash*/
    if (!in_node)
        return return_pos;
                  
    if (dec < 0)
        return *this + (-dec);

    return_pos = *this;

    if (0 == dec)
        return return_pos;

    if (char_index >= dec){
        return_pos.char_index -= dec;
        return return_pos;
        }
    
    i = dec - char_index - 1;

    return_pos.char_index = 0;
    return_pos = return_pos.get_prior();
    if (!return_pos.in_node)
        return *this;
    else if (0 == i)
        return return_pos;
                      
    hold_pos = return_pos;
    return_pos -= i;
    if (hold_pos == return_pos)
        return *this;
    else
        return return_pos;
}                                 

/* the - operator with model position argument*/
long model_pos::operator- (model_pos dec)
{
             model_pos hold_pos,            
                       first_pos,
                       last_pos;
             long      sign_multip;
    register long      return_val;
                         
    /*first, run a bunch of reality tests*/
    if (!this || !dec.in_node)
        return 0;
    if (*this == dec)
        return 0;
    if (in_node->get_model() != dec.in_node->get_model())
        return 0;
    if (!dec.in_node->can_format()){
        hold_pos = dec.get_next();
        if (!hold_pos.in_node)
            hold_pos = dec.get_prior();
        dec = hold_pos;
        }
    hold_pos = *this;
    if (!in_node->can_format()){
        hold_pos = hold_pos.get_next();
        if (!hold_pos.in_node)
            hold_pos = get_prior();
        }
    if (!hold_pos.in_node || !dec.in_node)
        return 0;

    /*we have something we can work with--now find out which is first*/
    if (hold_pos > dec){
        first_pos = dec;
        last_pos = hold_pos;
        sign_multip = 0;
        }
    else {
        first_pos = hold_pos;
        last_pos = dec;
        sign_multip = -1;
        }
 
    /*now find the number of characters between them*/
    if (first_pos.in_node == last_pos.in_node)
        return (last_pos.char_index - first_pos.char_index);

    return_val = first_pos.in_node->get_max_index() + 1 - first_pos.char_index;
    return_val += last_pos.char_index;
    first_pos.char_index = first_pos.in_node->get_max_index();
    first_pos = first_pos.get_next();
    while(first_pos.in_node != last_pos.in_node){
        first_pos.char_index = first_pos.in_node->get_max_index();
        return_val += first_pos.char_index + 1;
        first_pos = first_pos.get_next();
        }
    /*get the sign right*/
    if (sign_multip)
        return return_val * sign_multip;
    return return_val;
}

/*the ++ operator*/
model_pos model_pos::operator++()
{
    model_pos return_pos;
         
    if (char_index < in_node->get_max_index()){
        return_pos.in_node = in_node;
        return_pos.char_index++;
        return return_pos;
        }
    return_pos = get_next();
    if (!return_pos.in_node)
        return *this;
    else
        return return_pos;
}

/*the -- operator*/
model_pos model_pos::operator--()
{
    model_pos return_pos;
                         
    if (char_index > 0){
        return_pos.in_node = in_node;
        return_pos.char_index--;
        return return_pos;
        }
    return_pos = get_prior();
    if (!return_pos.in_node)
        return *this;
    else
        return return_pos;
}


/*get the next formattable model position*/
model_pos model_pos::get_next()
{
    model_pos return_pos;
    node_dir    dir;
    register node *return_node;

    /*make sure the model position is real; if not return it*/
    if ( !in_node)
        return *this;

    return_node = in_node;

    if (return_node->can_format() && (return_node->get_max_index() > char_index)){
        return_pos.in_node = in_node;
        return_pos.char_index = char_index + 1;
        return return_pos;
        }

    dir = n_down;
    for (return_node = return_node->get_next(dir);
           return_node;
           return_node = return_node->get_next(dir))
        if (return_node->can_format()) break;

    if (return_node)
        return_pos.char_index = 0;
    else
        return_pos.char_index =-1;
    return_pos.in_node = return_node;
    return return_pos;
}

/*get the previous character in the model*/
model_pos model_pos::get_prior()
{
    model_pos return_pos;
    node_dir    dir;
    register node      *return_node;

    /*make sure the model position is real; if not return it*/
    if ( !in_node)
        return *this;

    return_node = in_node;

    if (return_node->can_format() && (char_index > 0)) {
        return_pos.in_node = in_node;
        return_pos.char_index = char_index - 1;
        return return_pos;
        }

    dir = n_up;
    for (return_node = return_node->get_prev(dir);
            return_node;
            return_node = return_node->get_prev(dir))
        if (return_node->can_format()) break;

    if (return_node)
        return_pos.char_index = return_node->get_max_index();
    else
        return_pos.char_index = -1;
    return_pos.in_node = return_node;
    return return_pos;
}
 
/*can the model position be formatted?*/
long model_pos::can_format()
{
    return (in_node? in_node->can_format() : 0);
}

/*mark any pages containing a model position as needing reformatting*/
void model_pos::mark_changed()
{
    register mf_list   *format_list;
    register phead    *on_phead;

    if (!in_node || !in_node->get_model())
        return;

    for (format_list = in_node->get_model()->get_formats();
          format_list;
          format_list = format_list->inq_next()){
        on_phead = format_list->inq_format()->mpos_2_phead(*this);
        if (on_phead)
            on_phead->mark_reformat();
        }
}

/*redraw the line the model position is on if it is visible in any viewport. This routine assumes that
    the changes in the model will not change the size of the line drawn or the beginning and end
    points of the line*/
void model_pos::in_place_redraw()
{
    register mf_list   *format_list;
    format             *cur_format;
    register fv_list   *viewport_list;
    view_port          *cur_viewport;
    model_pos           start_redraw;
    register view_pos  *in_vpos,
                       *attach_vpos;
    register cline     *draw_cline;
    vp_dir              attach_dir;

    if (!in_node || !in_node->get_model())
        return;

    for (format_list = in_node->get_model()->get_formats();
            format_list;
            format_list = format_list->inq_next()){
        cur_format = format_list->inq_format();
        for (viewport_list = cur_format->get_fv_list();
                viewport_list;
                viewport_list = viewport_list->inq_next()){
            cur_viewport = viewport_list->inq_vp();
            (*cur_viewport->get_ftbl()->ptr_2_tml_dp_damage_repair)(cur_viewport->get_handle());
            if (in_vpos = cur_viewport->mpos_2_vpos(*this)){
                if (in_vpos->get_next()){
                    attach_dir = vp_up;
                    attach_vpos = in_vpos->get_next();
                    }
                else {
                    attach_dir = vp_down;
                    attach_vpos = in_vpos->get_prior();
                    }
                draw_cline = in_vpos->get_cline();
                start_redraw = draw_cline->get_start();
                if (*this < start_redraw)
                    start_redraw = *this;
                cur_viewport->redraw(start_redraw, *this, in_vpos);
                }
            }
        }
}

