/**************************************************************************************
    node_func.cxx
    functions built from the node traits and operating on the nodes of a
      structure tree in a text management library model

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_os_depends.hxx>

/*  node_alloc
allocate a node in a given model--default to allocating in the current model*/
node* node_alloc(long node_size, model* in_model)
{
    register node* new_node;

    new_node = in_model ? in_model->make_node(node_size) : (node*)(new char[node_size]);
    return(new_node);
}

                                                                                                              
/*  pvec_alloc
allocate a property vector in the indicated model--default to
allocating in the current model*/
prop_vec* pvec_alloc(model* to_model)
{
    register prop_vec *new_pvec;

    new_pvec = to_model ? to_model->make_pvec() : new prop_vec;
    return(new_pvec);
}

/*  prop_alloc
allocate a property in the indicated model--default to allocating in the current model*/
property* prop_alloc(model* to_model)
{
    register property *new_prop;

    new_prop = to_model ? to_model->make_prop() : new property;
    return(new_prop);
}

/*   arg_alloc
allocate an argument vector element in the indicated model--default to
allocating in the current model*/
argument* arg_alloc(model* to_model)
{
    register argument *new_arg;

    new_arg = to_model ? to_model->make_arg((long) 0) : new int_arg;
    return(new_arg);
}


/*  fix_start - called by make_subtree*/
node * fix_start(node *start_node)
{
    register node    *new_parent;

    if (start_node->get_node(n_left)) {    //need to create a new parent delimiting this node
        new_parent = start_node->get_node(n_up)->split(start_node->get_sib_order());
        }
    else
        new_parent = start_node->get_node(n_up);

     return(new_parent);
}

/*  fix_end - called by make_subtree*/
node * fix_end(node *end_node)
{
    register node    *new_parent;

   if (end_node->get_node(n_right)) {    //need to create a new parent delimiting this node
        new_parent = end_node->get_node(n_up)->split(end_node->get_sib_order() + 1);
        }
    else
        new_parent = end_node->get_node(n_up);

     return(end_node->get_node(n_up));
}
         

/*  make_subtree
Given a start_node and an end_node, isolate a coherent
subtree subsuming all text from start to end, and creating any
nodes necessary to preserve the properties as far up the tree
as necessary.
  assumption: nodes provided are whole leaf nodes
  returns:    root of isolated subtree
*/

node* make_subtree(node* start_node, node* end_node)
{
   long        st_level,
              end_level;
   node       *new_root,
                *hold_node,
                *next_node;

   /*We need to arrange the tree so that the start and end nodes
   are at the same level (distance from the root).*/
                            
    st_level = start_node->get_level();
    end_level = end_node->get_level();
   while(st_level != end_level){
      /*leave this loop when st_level = end_level*/
      if (st_level > end_level){
         start_node = fix_start(start_node);
         st_level = start_node->get_level();
         }
      else {
         end_node = fix_end(end_node);
         end_level = end_node->get_level();
         }
      }

   /*Levels of starting point and ending point are now equal*/

   while(start_node != end_node && (start_node->get_level() > 1)){
      /*Leave this loop when either:
          1. start_node = end_node
          2. the parent of start_node is the root node*/
      start_node = fix_start(start_node);
      end_node = fix_end(end_node);
      }

   if ((start_node != end_node) && (start_node->get_level() == 1)){
      /*The parent of our starting point is the ROOT of the
      entire tree. We need to create an intervening node to
      serve as the root of our subtree...*/
      new_root = new bnode(start_node->get_model());
      start_node->connect_node(new_root, n_up);
      for (hold_node = end_node; hold_node != new_root; hold_node = next_node){
        next_node = hold_node->get_node(n_left);
        hold_node->move(start_node, n_right);
        }
        return new_root;
      }

    return start_node;
}


                                                    
/*   get_next_leaf  - Gets the next LEAF node*/
node* get_next_leaf(node* start_node)
{
   node_dir  dir;

   dir = n_down;
   for (start_node = start_node->get_next(dir);
        start_node; start_node = start_node->get_next(dir))
           if (start_node->get_type() == leaf) return(start_node);
   if (!start_node) return(0);
}



/*   get_prev_leaf  -  Gets the previous LEAF node*/
node* get_prev_leaf(node* start_node)
{
   node_dir  dir;

   dir = n_down;
   for (start_node = start_node->get_prev(dir);
        start_node; start_node = start_node->get_prev(dir))
           if (start_node->get_type() == leaf) return(start_node);
  if (!start_node) return(0);
}



/*   copy component
Copy text, from start point to end point, to a different model; start point
and end point each consist of a node and character index. All subtree structure and
properties are preserved at the target. The target model may be another
text file being edited, or it may be a 'paste buffer.'
    returns: root of the copied subtree in the target model.
Note: 'copy_component' does not connect the copied subtree into the target model;
you must call connect_node to do this.*/

node* copy_component(model_pos *start, model_pos *end, model *to_model)
{
   register node      *new_root,
                           *start_node,
                           *end_node;

    if (start->in_node->get_model() != end->in_node->get_model())
        return 0;

    if (start->char_index > 0)
        start_node = start->in_node->split(start->char_index);
    else
        start_node = start->in_node;
    if (end->in_node->get_max_index() < (end->char_index + 1)) {
        end_node = end->in_node->split(end->char_index + 1);
        end_node = end->in_node;
        }
    if (!start_node || !end_node) return 0;

    new_root = make_subtree(start_node, end_node);
    if (!new_root) return 0;
    return (new_root->copy_all(to_model));
}
/*create and return a pointer to a word wrap property, with the value indicated, in the given model*/
property* make_wrap_prop(long wrap_on, model *in_model)
{
    register argument    *new_arg;
    register property    *new_prop;

    if (!in_model) return 0;
    new_arg = in_model->make_arg(wrap_on);
    if (!new_arg) return 0;
    new_prop = in_model->make_prop();
    if (!new_prop) {
        delete new_arg;
        return 0;
        }
    new_prop->attach_entry(new_arg);
    new_prop->set_index(tml_wrap_prop);
    return new_prop;
}

/*create and return a pointer to a margin property, with the values indicated, in the given model*/
property* make_margin_prop(long left,
                                  long right,
                                  long top,
                                  long bottom,
                                  model *in_model)
{
    register property    *new_prop;
    register argument    *new_arg,
                                *next_arg;

    if (in_model){
        new_prop = in_model->make_prop();
        if (!new_prop) return 0;
        new_arg = in_model->make_arg(top);
        if (!new_arg) return (0);
        next_arg = in_model->make_arg(bottom);
        if (!next_arg){
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        next_arg = in_model->make_arg(left);
        if (!next_arg){
            delete new_arg->get_next();
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        next_arg = in_model->make_arg(right);
        if (!next_arg){
            delete new_arg->get_next()->get_next();
            delete new_arg->get_next();
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        }
    else {
        new_prop = new property;
        if (!new_prop) return 0;
        new_arg = new int_arg;
        if (new_arg) ((int_arg*)new_arg)->set_val(top);
        if (!new_arg) return (0);
        next_arg = new int_arg;
        if (next_arg) ((int_arg*)next_arg)->set_val(bottom);
        if (!next_arg){
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        next_arg = new int_arg;
        if (next_arg) ((int_arg*)next_arg)->set_val(left);
        if (!next_arg){
            delete new_arg->get_next();
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        next_arg = new int_arg;
        if (next_arg) ((int_arg*)next_arg)->set_val(right);
        if (!next_arg){
            delete new_arg->get_next()->get_next();
            delete new_arg->get_next();
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        }

    new_prop->set_index(tml_margin_prop);
    new_prop->attach_entry(new_arg);
    return new_prop;
}

/*create and return a pointer to a page size property, with the values indicated, in the given model*/
property* make_page_size_prop(long width, long depth, model *in_model)
{
    register property    *new_prop;
    register argument    *new_arg,
                                *next_arg;

    if (in_model)
        new_arg = in_model->make_arg(width);
    else {
        new_arg = new int_arg;
        if (new_arg) ((int_arg*)new_arg)->set_val(width);
        }
    if (!new_arg) return 0;

    if (in_model)
        next_arg = in_model->make_arg(depth);
    else {
        next_arg = new int_arg;
        if (next_arg) ((int_arg*)next_arg)->set_val(depth);
        }
    if (!next_arg) {
        delete new_arg;
        return 0;
        }

    new_prop = in_model ? in_model->make_prop() : new property;
    if (!new_prop) {
        delete new_arg;
        delete next_arg;
        return 0;
        }
                               
    new_prop->set_index(tml_pgsize_prop);
    new_arg->append(next_arg);
    new_prop->attach_entry(new_arg);
    return new_prop;
}

/* create and return a pointer to a font property, with values indicated, in the given model */
property *make_font_prop(char *font_name, long name_len, model *in_model, grafsys for_grafsys)
{
    register argument      *new_arg;
    register argument      *next_arg;
    register font_property *new_prop;

    extern void* font_tbl_set_sys_from_type();

    if (0 == for_grafsys)
        for_grafsys = tml_dp_get_dflt_grafsys_type();

    if (font_name[name_len -1] == '\0')
        name_len--;
    if (in_model)
        new_arg = in_model->make_arg(font_name, name_len);
    else {
        new_arg = new string_arg;
        if (new_arg) ((string_arg*)new_arg)->set_val(font_name, name_len);
        }
    if (!new_arg) return 0;

    if (in_model)
        next_arg = in_model->make_arg((long) for_grafsys);
    else {
        next_arg = new int_arg;
        if (next_arg) 
            ((int_arg*)next_arg)->set_val((long) for_grafsys);
        }
    if (!next_arg) {
        delete new_arg;
        return 0;
        }

    if (in_model)
        new_prop = (font_property*) in_model->make_prop((long) sizeof(font_property));
    else
        new_prop = new font_property;
    if (!new_prop) {
        delete new_arg;
        delete next_arg->get_next();
        delete next_arg;
        return 0;
        }

    new_prop->set_ftbl((vpftbl*) for_grafsys->instance);
    new_prop->attach_entry(new_arg);
    new_arg->append(next_arg);
    new_prop->set_index(tml_font_prop);
    return new_prop;
}

/*create and return a pointer to an invert property, with the value indicated, in the given model*/
property* make_inv_prop(long inv_on, model *in_model)
{
    register argument    *new_arg;
    register property    *new_prop;

    if (in_model)
        new_arg = in_model->make_arg(inv_on);
    else {
        new_arg = new int_arg;
        if (new_arg) ((int_arg*)new_arg)->set_val(inv_on);
        }
    if (!new_arg) return 0;
    if (in_model)
        new_prop = in_model->make_prop();
    else
        new_prop = new property;
    if (!new_prop) {
        delete new_arg;
        return 0;
        }
    new_prop->attach_entry(new_arg);
    new_prop->set_index(tml_inv_prop);
    return new_prop;
}


/* create and return a pointer to a header property */
property* make_header_prop(model *in_model,
                           node  *left_hdr,
                           node  *center_hdr,
                           node  *right_hdr)
{
    register argument *new_arg;
    register argument *next_arg;
    register property *new_prop;

    if (in_model) {
        new_prop = in_model->make_prop();
        if (!new_prop)
            return 0;
        new_arg = in_model->make_arg(left_hdr);
        if (!new_arg) {
            delete new_prop;
            return 0;
            }
        next_arg = in_model->make_arg(center_hdr);
        if (!next_arg) {
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        next_arg = in_model->make_arg(right_hdr);
        if (!next_arg) {
            delete new_arg->get_next();
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        }
    else {
        new_prop = new property;
        if (!new_prop)
            return 0;
        new_arg = new subtree_arg;
        if (!new_arg) {
            delete new_prop;
            return 0;
            }
        ((subtree_arg*)new_arg)->set_val (left_hdr);
        next_arg = new subtree_arg;
        if (!next_arg) {
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else {
            ((subtree_arg*)next_arg)->set_val (center_hdr);
            new_arg->append(next_arg);
            }
        next_arg = new subtree_arg;
        if (!next_arg) {
            delete new_arg->get_next();
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else {
            ((subtree_arg*)next_arg)->set_val (right_hdr);
            new_arg->append(next_arg);
            }
        }
    new_prop->attach_entry(new_arg);
    new_prop->set_index(tml_header_prop);
    return new_prop;
}

/* create and return a pointer to a footer property */
property* make_footer_prop(model *in_model,
                           node  *left_ftr,
                           node  *center_ftr,
                           node  *right_ftr)
{
    register argument *new_arg;
    register argument *next_arg;
    register property *new_prop;

    if (in_model) {
        new_prop = in_model->make_prop();
        if (!new_prop)
            return 0;
        new_arg = in_model->make_arg(left_ftr);
        if (!new_arg) {
            delete new_prop;
            return 0;
            }
        next_arg = in_model->make_arg(center_ftr);
        if (!next_arg) {
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        next_arg = in_model->make_arg(right_ftr);
        if (!next_arg) {
            delete new_arg->get_next();
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        }
    else {
        new_prop = new property;
        if (!new_prop)
            return 0;
        new_arg = new subtree_arg;
        if (!new_arg) {
            delete new_prop;
            return 0;
            }
        ((subtree_arg*)new_arg)->set_val (left_ftr);
        next_arg = new subtree_arg;
        if (!next_arg) {
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else {
            ((subtree_arg*)next_arg)->set_val (center_ftr);
            new_arg->append(next_arg);
            }
        next_arg = new subtree_arg;
        if (!next_arg) {
            delete new_arg->get_next();
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else {
            ((subtree_arg*)next_arg)->set_val (right_ftr);
            new_arg->append(next_arg);
            }
        }
    new_prop->attach_entry(new_arg);
    new_prop->set_index(tml_footer_prop);
    return new_prop;
}

/*create and return a pointer to a form feed property, with the value indicated, in the given model*/
property* make_form_feed_prop(long ff_on, model *in_model)
{
    register argument    *new_arg;
    register property    *new_prop;

    if (!in_model) return 0;
    new_arg = in_model->make_arg(ff_on);
    if (!new_arg) return 0;
    new_prop = in_model->make_prop();
    if (!new_prop) {
        delete new_arg;
        return 0;
        }
    new_prop->attach_entry(new_arg);
    new_prop->set_index(tml_form_feed_prop);
    return new_prop;
}

/*create and return a pointer to a character set property, with the value indicated, in the given model*/
property* make_char_set_prop(char_set* cur_char_set, model *in_model)
{
    register argument    *new_arg;
    register property    *new_prop;

    if (!in_model)
        return 0;
    new_arg = in_model->make_arg(cur_char_set);
    if (!new_arg) return 0;
    new_prop = in_model ? in_model->make_prop() : new property;
    if (!new_prop) {
        delete new_arg;
        return 0;
        }
    new_prop->attach_entry(new_arg);
    new_prop->set_index(tml_char_set_prop);
    return new_prop;
}

/*create and return a pointer to a union character set property, with the value indicated, in the given model*/
property* make_union_char_set_prop(union_char_set* cur_union, model *in_model)
{
    register argument    *new_arg;
    register property    *new_prop;

    if (!in_model)
        return 0;
    new_arg = in_model->make_arg(cur_union);
    if (!new_arg) return 0;
    new_prop = in_model ? in_model->make_prop() : new property;
    if (!new_prop) {
        delete new_arg;
        return 0;
        }
    new_prop->attach_entry(new_arg);
    new_prop->set_index(tml_union_char_set_prop);
    return new_prop;
}

static void kill_arg_list(argument *arg_list)
{
  argument *next_arg;

  next_arg = arg_list->get_next();
  if(next_arg != NULL) kill_arg_list(next_arg);
  delete arg_list;
}

/* create and return a pointer to a tab property, a linked list of tab stops, in the given model */
property* make_tab_prop(model *in_mod, long default_val, long num_stops, long *tab_st_list)
{
    register argument *new_arg;
    register argument *next_arg;
    register int_arg  *new_iarg;
    register int_arg  *next_iarg;
    register property *new_prop;
    register long      i;
    register long     *tabval;

    if (in_mod){
        new_prop = in_mod->make_prop();
        if (!new_prop) return 0;
        new_arg = in_mod->make_arg(default_val);
        if (!new_arg) return 0;
        next_arg = in_mod->make_arg(num_stops);
        if (!next_arg){
            delete new_arg;
            delete new_prop;
            return 0;
            }
        else
            new_arg->append(next_arg);
        for(i = 0, tabval = tab_st_list; i < num_stops; i++, tabval++) {
          next_arg = in_mod->make_arg(*tabval);
          if(!next_arg) {
            kill_arg_list(new_arg);
            delete new_prop;
            return 0;
          }
          new_arg->append(next_arg);
        }
    }
    else {
        new_prop = new property;
        if (!new_prop) return 0;
        new_iarg = new int_arg;
        if (!new_iarg) return 0;
        new_arg = new_iarg;
        new_iarg->set_val(default_val);
        next_iarg = new int_arg;
        if (!next_iarg){
            delete new_iarg;
            delete new_prop;
            return 0;
            }
        else {
            next_iarg->set_val(num_stops);
            new_iarg->append(next_iarg);
        }
        for(i = 0, tabval = tab_st_list; i < num_stops; i++, tabval++) {
          next_iarg = new int_arg;
          if(!next_iarg) {
            kill_arg_list(new_iarg);
            delete new_prop;
            return 0;
          }
          next_iarg->set_val(*tabval);
          new_iarg->append(next_iarg);
       }
    }
    new_prop->set_index(tml_tab_prop);
    new_prop->attach_entry(new_arg);
    new_prop->attach_exit(new_arg);
    return new_prop;
}

/*connect a property to a node at the end of that node's property vector*/
void connect_prop(property *add_prop, node *to_node)
{
    register property    *new_prop;
    register prop_vec   *new_vec;

    if (add_prop->get_model() != to_node->get_model())
        new_prop = add_prop->copy(to_node->get_model());
    else
        new_prop = add_prop;

    new_vec = pvec_alloc(to_node->get_model());
    new_vec->attach(new_prop);
    if (to_node->get_props())
        to_node->get_props()->append(new_vec);
    else
        to_node->prop_attach(new_vec);
}

overload read_uasc;
model* read_uasc(void*);
model* read_uasc(char*, long);

/* return a pointer to a model whose nodes are filled with the characters in the */
/*   uasc file specified by pathname and name length */
model* read_uasc(char *pathname, long name_len)
{
    void *fdp;
    model *new_model;

    if (NULL == (fdp = tml_os_open(pathname, name_len, O_RDONLY, 0))) {
        printf("read_uasc(char *pathname, long name_len): tml_os_open failed\n");
        return(NULL);
        }
    new_model = read_uasc(fdp);
    if (-1 == tml_os_close(fdp))
        printf("read_uasc(char *pathname, long name_len): tml_os_close failed\n");
    return(new_model);
}

/* return a pointer to a model whose nodes are filled with the characters in the */
/*   uasc file indicated by the file handle pointer returned by tml_os_open() */
model* read_uasc(void *fdp)
{
    char inbuf[16000];

    model *new_model;
    node *cur_node;
    gnode *new_node;
    unsigned long to_read;
    long r;

    new_model = new model;
    cur_node = new_model->make_root();
    to_read = (16000 <= new_model->get_gheap_size()) ? 16000 : new_model->get_gheap_size();
    while (r = tml_os_read(fdp, inbuf, to_read)) {
        if (-1 == r) {
            printf("read_uasc: read failed\n");
            delete new_model;
            return(NULL);
            }
        else {
            new_node = new gnode(new_model);
            if (NULL != cur_node->get_node(n_up))
                new_node = (gnode *) cur_node->connect_node(new_node, n_right);
            else
                new_node = (gnode *) cur_node->connect_node(new_node, n_down);
            if (!new_node->attach_data(inbuf, r))
                printf("unable to attach %d bytes of data\n", r);
            else
                cur_node = (node *) new_node;
            }
        }
    return(new_model);
}
 

long write_uasc(model*, unsigned char*, long)
{
    return 0;
}


/*this function exists only so the libcxx function _entry() will
    be bound in, allowing reference by the c-interface function
    tml_init()*/
void tml_bogus_entry()
{
#ifndef hpux
// hpux uses _main() at C++ 1.2
#ifndef __cplusplus
// suitable for Glockenspiel C++ at 1.2 and earlier
    _entry();
#else 
// suitable for AT&T C++
    _main();
#endif
#else
// suitable for AT&T C++
    _main();
#endif
}

/*character-set-dependent string functions using default character set for model*/
char* strcat_for_model (model* in_model, char* s1, char* s2)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return ((char*) ((*dflt_cs->strcat_p)((unsigned char*)s1, (unsigned char*)s2, dflt_cs)));
}
        
char* strncat_for_model(model* in_model, char* s1, char* s2, long n)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return ((char*) ((*dflt_cs->strncat_p)((unsigned char*)s1, (unsigned char*)s2, n, dflt_cs)));
}

long  strcmp_for_model (model* in_model, char* s1, char* s2)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return (*dflt_cs->strcmp_p)((unsigned char*)s1, (unsigned char*)s2, dflt_cs);
}
        
long  strncmp_for_model(model* in_model, char* s1, char* s2, long n)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return (*dflt_cs->strncmp_p)((unsigned char*)s1, (unsigned char*)s2, n, dflt_cs);
}

char* strcpy_for_model (model* in_model, char* s1, char* s2)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return ((char*) ((*dflt_cs->strcpy_p)((unsigned char*)s1, (unsigned char*)s2, dflt_cs)));
}
        
char* strncpy_for_model(model* in_model, char* s1, char* s2, long n)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return ((char*) ((*dflt_cs->strncpy_p)((unsigned char*)s1, (unsigned char*)s2, n, dflt_cs)));
}

long  strlen_for_model (model* in_model, char* s)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return (*dflt_cs->strlen_p)((unsigned char*)s, dflt_cs);
}
                     
char* index_for_model  (model* in_model, char* s, char* c)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return ((char*) ((*dflt_cs->index_p)((unsigned char*)s, (unsigned char*)c, dflt_cs)));
}
    
char* rindex_for_model (model* in_model, char* s, char* c)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return ((char*) ((*dflt_cs->rindex_p)((unsigned char*)s, (unsigned char*)c, dflt_cs)));
}

/*character-set-dependent test for null*/
long  is_null_for_model(model* in_model, char* c)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return dflt_cs->is_null((unsigned char*) c);
}

/*character-set-dependent COLLATING CHARACTER string comparisons*/
long char_arr_cmp_for_model(model* in_model, char* s1, char* s2)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return dflt_cs->char_arr_cmp((unsigned char*)s1, (unsigned char*)s2);
}

long char_arr_ncmp_model(model* in_model, char* s1, char* s2, unsigned long n)
{
    char_set* dflt_cs;

    dflt_cs = in_model->get_char_sets()->get_sets();
    return dflt_cs->char_arr_ncmp((unsigned char*)s1,(unsigned char*) s2, n);
}

/*declarations of character-set-dependent string functions using character set in effect at a node*/
char* strcat_for_node (node* at_node, char* s1, char* s2)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return ((char*) ((*cur_cs->strcat_p)((unsigned char*)s1, (unsigned char*)s2, cur_cs)));
}
        
char* strncat_for_node(node* at_node, char* s1, char* s2, long n)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return ((char*) ((*cur_cs->strncat_p)((unsigned char*)s1, (unsigned char*)s2, n, cur_cs)));
}

long  strcmp_for_node (node* at_node, char* s1, char* s2)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return (*cur_cs->strcmp_p)((unsigned char*)s1, (unsigned char*)s2, cur_cs);
}
        
long  strncmp_for_node(node* at_node, char* s1, char* s2, long n)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return (*cur_cs->strncmp_p)((unsigned char*)s1, (unsigned char*)s2, n, cur_cs);
}

char* strcpy_for_node (node* at_node, char* s1, char* s2)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return ((char*) ((*cur_cs->strcpy_p)((unsigned char*)s1, (unsigned char*)s2, cur_cs)));
}
        
char* strncpy_for_node(node* at_node, char* s1, char* s2, long n)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return ((char*) ((*cur_cs->strncpy_p)((unsigned char*)s1, (unsigned char*)s2, n, cur_cs)));
}

long  strlen_for_node (node* at_node, char* s)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return (*cur_cs->strlen_p)((unsigned char*)s, cur_cs);
}
                     
char* index_for_node  (node* at_node, char* s, char* c)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return ((char*) ((*cur_cs->index_p)((unsigned char*)s, (unsigned char*)c, cur_cs)));
}
     
char* rindex_for_node (node* at_node, char* s, char* c)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return ((char*) ((*cur_cs->rindex_p)((unsigned char*)s, (unsigned char*)c, cur_cs)));
}

/*character-set-dependent test for null*/
long  is_null_for_node(node* at_node, char* c)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return cur_cs->is_null((unsigned char*)c);
}

/*character-set-dependent COLLATING CHARACTER string comparisons*/
long char_arr_cmp_for_node(node* at_node, char* s1, char* s2)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return cur_cs->char_arr_cmp((unsigned char*)s1, (unsigned char*)s2);
}

long char_arr_ncmp_for_node(node* at_node, char* s1, char* s2, unsigned long n)
{
    char_set* cur_cs;

    cur_cs = (char_set*) at_node->get_model()->inq_state(at_node, tml_char_set_prop);
    return cur_cs->char_arr_ncmp((unsigned char*)s1, (unsigned char*)s2, n);
}

