/**************************************************************************************
    tml_os_sys5.cxx
    definitions of the operating system dependent routines in tml++
      for UNIX sys5.3
    
    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <signal.h>
#include <errno.h> 
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <memory.h>
#ifdef __cplusplus
#include <stdlib.h>
#include <osfcn.h>
#endif
#include <tml/arch_neutral.hxx>      /* defines WHOLE_PTR */
#include <tml/csm_lock.hxx>
#include <tml/tml_os_depends.hxx>

/* copy count bytes at from address into space beginning at to address */
char* tml_os_byte_copy(char* from, char* to, long count)
{
    char* end_addr;

    /*check for incoming trash*/
    if (!to || !from)
        return 0;
    /*make sure we don't overwrite*/
    if ((from < to) && ((from + count) > to))
        return 0;

    end_addr = (char*) (memcpy(to, from, (int) count));
    return(end_addr);
}

/* a unique overloaded definition for this function must be included */
/* for each type of comparison function */
void tml_os_qsort( char *base, unsigned long nel, unsigned long size, long (*cmp)(void*, void*) )
{
#ifdef __cplusplus
    typedef int (*PFIVV)(const void*, const void*);
    PFIVV cast_cmp;
#else
    extern void qsort(char *, unsigned long, unsigned long, long (*)(...));
    PFI  cast_cmp;
#endif

    /*check for incoming trash*/
    if (!base || !nel || !size || !cmp)
        return;

#ifdef __cplusplus
    cast_cmp = (PFIVV) cmp;
    qsort(base, nel, size, cast_cmp);
#else
    cast_cmp = (PFI) cmp;
    qsort(base, nel, size, cast_cmp);
#endif
}

/* return NULL on failure or pointer to nelem * elsize bytes of free space. */
/*   Each byte is initialized to zero. */
char* tml_os_alloc(unsigned long nelem, unsigned long elsize)
{
    extern char* tml_c_new_wrapper(unsigned long num_bytes);

    register char* temp;
    register long  i;
    register long  quantity;

    quantity = nelem * elsize;
    temp = new char[quantity];
    for (i = 0; i < quantity; i++)
        *(temp + i) = 0;
    return temp;
}

/* free the space at ptr, which must have been obtained from a call to */
/*   tml_os_alloc(), or the result is unpredictible. */
void tml_os_free(char* ptr)
{
    delete ptr;
}

/* Pointer comparison function for calls to tml_os_qsort() in d_write_model() that sort */
/*   the disk model information structure's arrays of heap pointers */
long tml_os_ptr_cmp(char* *ptr1, char* *ptr2)
{
    if ((WHOLE_PTR) *ptr1 > (WHOLE_PTR) *ptr2)
        return(1);
    else if ((WHOLE_PTR) *ptr1 < (WHOLE_PTR) *ptr2)
        return(-1);
    else
        return(0);
}

/* Allocate and initialize a semaphore structure with a given capacity */
void* tml_os_init_sem(long pool_capacity)
{
    long sem_id;
#ifdef apollo 
#ifndef __cplusplus
    int  arg;
#else
    union semun arg;
#endif
#else
    union semun arg;
#endif

    /* key value IPC_PRIVATE guarantees return of unused entry */
    if (-1 == (sem_id = semget(IPC_PRIVATE, 1, IPC_CREAT))) {
        fprintf(stderr, "tml_os_init_sem: semget failed\n");
        return NULL;
        }

#ifdef apollo
#ifndef __cplusplus
    arg = pool_capacity;
#else
    arg.val = (int) pool_capacity;
#endif
#else
    arg.val = (int) pool_capacity;
#endif

    if (-1 == (semctl((int) sem_id, 0, SETVAL, arg))) {
        fprintf(stderr, "tml_os_init_sem: semctl initialization failed\n");
        if (-1 == (semctl((int) sem_id, 0, IPC_RMID, arg)))
            fprintf(stderr, "tml_os_init_sem: semctl deletion failed\n");
        return NULL;
        }
    return (void*) sem_id;
}

/* Deallocate a semaphore structure */
void tml_os_free_sem(void* old_sem)
{
    long sem_id;
#ifdef apollo
#ifndef __cplusplus
    int  arg;
#else
    union semun arg;
#endif
#else
    union semun arg;
#endif

    sem_id = (long) old_sem;

    if (-1 == (semctl((int) sem_id, 0, IPC_RMID, arg))) {
        fprintf(stderr, "tml_os_free_sem: semctl deletion failed\n");
        return;
        }
}

/* Busy wait on a semaphore */
long tml_os_busy_wait(void* on_sem)
{
    long sem_id;
#ifdef apollo
#ifndef __cplusplus
    int  arg;
#else
    union semun arg;
#endif
#else
    union semun arg;
#endif

    sem_id = (long) on_sem;

    if (0 >= (semctl((int) sem_id, 0, GETVAL, arg)))
        return -1;
    return 0;
}

/* Lock a semaphore */
void tml_os_p_sem(void* on_sem)
{
           void*  wassig;
           long   sem_id;
    struct sembuf sop;

    // record current interrupt handling status
    //   and then ignore interrupts
    wassig = tml_os_ignore_interrupts();

    sem_id = (long) on_sem;
    sop.sem_num = 0;
    sop.sem_op = -1;
    sop.sem_flg = SEM_UNDO;
    if (-1 == (semop((int) sem_id, &sop, 1)))
        fprintf(stderr, "tml_os_p_sem: semop failed\n");

    // restore interrupt handling on wakeup
    //   or immediately if semaphore was unlocked
    wassig = tml_os_restore_interrupts(wassig);
}

/* Continue all processes waiting on semaphore queue */
void* tml_os_v_sem(void* proc_id, void* tml_sem)
{
    void*          wassig;
    long           sem_id;
    struct sembuf  sop;
    void*          retval;

    // lock out interrupts while releasing lock
    wassig = tml_os_ignore_interrupts();

    sem_id = (long) tml_sem;
    sop.sem_num = 0;
    sop.sem_op = 1;
    sop.sem_flg = SEM_UNDO;
    if (-1 == (semop((int) sem_id, &sop, 1))) {
        fprintf(stderr, "tml_os_v_sem: semop failed\n");
        retval = (void*) -1;
        }
    else
        retval = (void*) 0;
    
    // restore interrupts after lock processing complete
    wassig = tml_os_restore_interrupts(wassig);
    
    return retval;
}

/* Check availability of a resource via its semaphore */
long tml_os_resource_available(void* on_sem)
{
    long sem_id;
#ifdef apollo
#ifndef __cplusplus
    int  arg;
#else
    union semun arg;
#endif
#else
    union semun arg;
#endif

    sem_id = (long) on_sem;
    if (0 < (semctl((int) sem_id, 0, GETVAL, arg)))
        return -1;
    return 0;
}

/* Return process identifier for the current process -- return value is */
/*   a long value cast as a void* */
void* tml_os_getpid()
{
    return (void*) getpid();
}

/* Reduce vulnerability of the current process to interrupts during critical */
/*   code sections -- returns status of interrupts before the call */
void* tml_os_ignore_interrupts()
{
    return (void*) signal(SIGINT, SIG_IGN);
}

/* Restore vulnerability of the current process to interrupts after completing */
/*   critical code sections -- returns status of interrupts after the call */
void* tml_os_restore_interrupts(void* interrupts)
{
#ifdef apollo 
#ifndef __cplusplus
    typedef void (*PFV)();

    return (void*) signal(SIGINT, (PFV) interrupts);
#else
    return (void*) signal(SIGINT, (SIG_PF) interrupts);
#endif
#else
    return (void*) signal(SIGINT, (SIG_PF) interrupts);
#endif

}

/* Kill a process dead -- returns a pointer to the appropriate object */
/*   cast as a void* */
void* tml_os_kill_proc(void* proc_id)
{
    if (proc_id)
        return (void*) kill((int) proc_id, SIGKILL);
    else
        return (void*) NULL;
}

/* Argument "file_handle" is the target of the write operation, "bufs" is a set of buffers, */
/*   "nbyte_arr" is their lengths in bytes, "sems" is a semaphore for each buffer. */
/* Do not wait for write operations to complete. */
/* The buffers are guaranteed to be written in the order given */
/*   except when a pending call to the same function accesses */
/*   the same semaphores. */
/* The semaphores are the tml_sem members of csm_lock objects corresponding to the buffers. */
long tml_os_asynch_write(void* file_handle, long num_bufs, char** bufs, long* nbyte_arr, csm_lock** locks)
{
    long i;
    long bytes_written;
    long proc_id;

    if (0 == (proc_id = fork())) {
        /* in child */
        for (i = 0; i < num_bufs; i++) {
            (*(locks + i))->p_suspend();
            if (-1 == (tml_os_write(file_handle, *(bufs + i), *(nbyte_arr + i)))) {
                printf("tml_os_asynch_write: can't write\n");
                return -1;
                }
            (*(locks + i))->v();
            }
        }
    for (bytes_written = 0, i = 0; i < num_bufs; i++)
        bytes_written += *(nbyte_arr + i);
    return bytes_written;
}

